# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Compute and display a Pareto Front
**********************************
"""
from __future__ import division, unicode_literals

from itertools import combinations
from typing import Optional, Sequence, Tuple

import matplotlib
import matplotlib.pyplot as plt
from numpy import all as np_all
from numpy import any as np_any
from numpy import array, full, ndarray, vstack


def compute_pareto_optimal_points(
    obj_values,  # type: ndarray
    feasible_points=None,  # type: Optional[ndarray]
):  # type: (...) -> ndarray
    """Compute the Pareto optimal points.

    Search for all the non-dominated points, i.e. there exists ``j`` such that
    there is no lower value for ``obj_values[:,j]`` that does not degrade
    at least one other objective ``obj_values[:,i]``.

    Args:
        obj_values: The objective function array of size `(n_samples, n_objs)`.
        feasible_points: An array of boolean of size n_sample,
            True if the sample is feasible,
            False otherwise.

    Returns:
        The vector of booleans of size n_samples, True if
        the point is Pareto optimal.
    """
    pareto_optimal = full(obj_values.shape[0], True)

    if feasible_points is None:
        feasible_points = full(obj_values.shape[0], True)

    def any_ax1_all(arr):
        return np_all(np_any(arr, axis=1))

    # Store the feasible indexes
    feasible_indexes = []
    for i, feasible_point in enumerate(feasible_points):
        if not feasible_point:
            pareto_optimal[i] = False
        else:
            feasible_indexes.append(i)

    # Exclude the non-feasible points for the computation of the Pareto optimal points
    obj_values_filtered = obj_values[feasible_indexes, :]
    for i, feasible_index in enumerate(feasible_indexes):
        obj = obj_values[feasible_index]
        before_are_worse = any_ax1_all(obj_values_filtered[:i] > obj)
        after_are_worse = any_ax1_all(obj_values_filtered[i + 1 :] > obj)
        pareto_optimal[feasible_index] = before_are_worse and after_are_worse

    return pareto_optimal


class ParetoPlotBiObjective(object):
    """Plot a 2D Pareto front on a Matplotlib axes."""

    def __init__(
        self,
        axes,  # type: matplotlib.axes.Axes
        obj_values,  # type: ndarray
        pareto_optimal_loc,  # type: Sequence[bool]
        obj_names,  # type: Sequence[str]
        all_pareto_optimal,  # type: Sequence[bool]
        is_non_feasible,  # type: Sequence[bool]
        bi_obj=False,  # type: bool
        show_non_feasible=True,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            axes: A matplotlib axes on which to be plotted.
            obj_values: The objective function array, of size (n_samples, n_objs).
            pareto_optimal_loc: A vector of booleans of size n_samples,
                True if the point is Pareto optimal.
            obj_names: The names of the objectives.
            all_pareto_optimal: The indices of points that are Pareto optimal
                w.r.t. all criteria.
            is_non_feasible: An Array of booleans of size n_samples,
                True if non_feasible.
            bi_obj: True if there are only two objective values.
            show_non_feasible: True to show the non-feasible points.
        """
        self.__nb_pareto_pts = obj_values.shape[0]
        self.__axes = axes
        self.__obj_values = obj_values
        self.__pareto_optimal_loc = pareto_optimal_loc
        self.__obj_names = obj_names
        self.__pareto_optimal_all = all_pareto_optimal
        self.__is_non_feasible = is_non_feasible
        self.__bi_objective = bi_obj
        self.__show_non_feasible = show_non_feasible
        self.__pareto_dominated_indexes = None

        # Compute the Pareto dominated indexes
        self.__compute_pareto_dominated_indexes()

    def plot_on_axes(self):  # type: (...) -> None
        """Plot the Pareto points on the Matplolib axes."""
        self.__plot_pareto_dominated_points()
        self.__plot_non_feasible_points()
        self.__plot_globally_pareto_optimal_points()
        self.__plot_locally_pareto_optimal_points()
        self.__axes.set_xlabel(self.__obj_names[0])
        self.__axes.set_ylabel(self.__obj_names[1])

    def __compute_pareto_dominated_indexes(self):  # type: (...) -> None
        """Compute the Pareto dominated indexes.

        The Pareto-dominated points are all the points which are feasible, but not
        locally and globally pareto optimal.
        """
        pareto_dominated_indexes = array([True] * self.__nb_pareto_pts)
        self.__pareto_dominated_indexes = (
            pareto_dominated_indexes
            & ~array(self.__is_non_feasible)
            & ~self.__pareto_optimal_loc
            & ~self.__pareto_optimal_all
        )

    def __plot_pareto_dominated_points(self):  # type: (...) -> None
        """Plot the Pareto-dominated points on the scatter plot."""
        if True in self.__pareto_dominated_indexes:
            self.__axes.scatter(
                self.__obj_values[self.__pareto_dominated_indexes, 0],
                self.__obj_values[self.__pareto_dominated_indexes, 1],
                color="b",
                label="Pareto dominated",
            )

    def __plot_non_feasible_points(self):  # type: (...) -> None
        """Plot the non-feasible points on the scatter plot."""
        if True in self.__is_non_feasible and self.__show_non_feasible:
            self.__axes.scatter(
                self.__obj_values[self.__is_non_feasible, 0],
                self.__obj_values[self.__is_non_feasible, 1],
                color="r",
                marker="X",
                label="Non feasible point",
            )

    def __plot_globally_pareto_optimal_points(self):  # type: (...) -> None
        """Plot the globally optimal Pareto points on the scatter plot."""
        if True in self.__pareto_optimal_all:
            if self.__bi_objective:
                label = "Pareto optimal"
            else:
                label = "Globally Pareto optimal"
            self.__axes.scatter(
                self.__obj_values[self.__pareto_optimal_all, 0],
                self.__obj_values[self.__pareto_optimal_all, 1],
                color="g",
                label=label,
            )

    def __plot_locally_pareto_optimal_points(self):  # type: (...) -> None
        """Plot the locally optimal Pareto points on the scatter plot."""
        if True in self.__pareto_optimal_loc and not self.__bi_objective:
            self.__axes.scatter(
                self.__obj_values[self.__pareto_optimal_loc, 0],
                self.__obj_values[self.__pareto_optimal_loc, 1],
                color="r",
                label="Locally Pareto optimal",
            )


def generate_pareto_plots(
    obj_values,  # type: ndarray
    obj_names,  # type: Sequence[str]
    figsize=(10.0, 10.0),  # type: Tuple[float, float]
    non_feasible_samples=None,  # type: Optional[ndarray]
    show_non_feasible=True,  # type: bool
):  # type: (...) -> matplotlib.figure.Figure
    """Plot a 2D Pareto front.

    Args:
        obj_values: The objective function array of size (n_samples, n_objs).
        obj_names: The names of the objectives.
        figsize: The matplotlib figure sizes in x an y directions, in inches.
        non_feasible_samples: The array of bool of size n_samples,
            True if the current sample is non-feasible.
            If None, all the samples are considered feasible.
        show_non_feasible: If True, show the non-feasible points in
            the Pareto front plot.

    Raises:
        ValueError: If the number of objective values and names are different.
    """
    n_obj = obj_values.shape[1]
    nb_pareto_pts = obj_values.shape[0]
    obj_names_length = len(obj_names)
    if obj_names_length != n_obj:
        msg = "Inconsistent objective values size and objective names: {} != {}".format(
            n_obj, obj_names_length
        )
        raise ValueError(msg)

    # If non_feasible_samples set to None,
    # then all the points are considered as feasible
    if non_feasible_samples is None:
        non_feasible_samples = full(nb_pareto_pts, False)
    is_feasible = ~non_feasible_samples

    pareto_opt_all = compute_pareto_optimal_points(obj_values, is_feasible)

    fig, axes = plt.subplots(n_obj - 1, n_obj - 1, figsize=figsize, squeeze=False)
    fig.suptitle("Pareto front")

    # 0 vs 1   0 vs 2    0 vs 3
    #          1 vs 2    1 vs 3
    #                    2 vs 3

    # i j+1     i=0...nobj-1
    #           j=1....nobj

    for i, j in combinations(range(n_obj), 2):  # no duplication, j!=j
        obj_loc = vstack((obj_values[:, i], obj_values[:, j])).T
        pareto_opt_loc = compute_pareto_optimal_points(obj_loc, is_feasible)

        axe = axes[i, j - 1]
        bi_obj = True if n_obj == 2 else False
        plot_pareto_bi_obj = ParetoPlotBiObjective(
            axe,
            obj_loc,
            pareto_opt_loc,
            [obj_names[i], obj_names[j]],
            pareto_opt_all,
            non_feasible_samples,
            bi_obj=bi_obj,
            show_non_feasible=show_non_feasible,
        )
        plot_pareto_bi_obj.plot_on_axes()

        if i != j - 1:
            axes[j - 1, i].remove()

    # Ensure the unicity of the labels in the legend
    new_handles = []
    new_labels = []
    for axe in axes.flatten():
        handles, labels = axe.get_legend_handles_labels()
        for label, handle in zip(labels, handles):
            if label not in new_labels:
                new_labels.append(label)
                new_handles.append(handle)
    fig.legend(new_handles, new_labels, loc="lower left")

    return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard, Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Linear solvers wrapper
**********************
"""
from __future__ import division, unicode_literals

import logging

import numpy as np
import scipy.sparse.linalg as scipy_linalg
from scipy.sparse.base import issparse
from scipy.sparse.linalg import bicgstab, cgs

LOGGER = logging.getLogger(__name__)


class LinearSolver(object):
    """Solve a linear system Ax=b."""

    LGMRES = "lgmres"
    AVAILABLE_SOLVERS = {LGMRES: scipy_linalg.lgmres}

    def __init__(self):
        """Constructor."""
        self.outer_v = []  # Used to store (v,Av) pairs for restart and multiple RHS

    @staticmethod
    def _check_linear_solver(linear_solver):
        """Check that linear solver is available.

        :param linear_solver: name of linear solver used to solve linear solver
        """
        solver = LinearSolver.AVAILABLE_SOLVERS.get(linear_solver, None)
        if solver is None:
            raise AttributeError(
                "Invalid linear solver" + str(linear_solver) + " for scipy sparse: "
            )
        return solver

    @staticmethod
    def _check_b(a_mat, b_vec):
        """Check the dimensions of the vector b converts it to ndarray if sparse, for
        lgmres needs.

        :param a_mat: the matrix A
        :param b_vec: the vector b
        :returns: the vector b with consistent dimensions
        """
        if len(b_vec.shape) == 2 and b_vec.shape[1] != 1:
            LOGGER.error(
                "Incompatible dimensions in linear system Ax=b, A "
                "shape is %s and b shape is %s",
                str(a_mat.shape),
                str(b_vec.shape),
            )
            raise ValueError(
                "Second member of the linear system" + " must be a column vector"
            )
        if issparse(b_vec):
            b_vec = b_vec.toarray()
        return b_vec.real

    def solve(self, a_mat, b_vec, linear_solver="lgmres", **kwargs_lin):
        """Solves the linear system Ax=b using scipy sparse GMRES solver.

        :param a_mat: matrix A of the system, can be a sparse matrix
        :param b_vec: second member
        :param linear_solver: name of linear solver (Default value = 'lgmres')
        :param kwargs_lin: arguments passed to the scipy linear solver
        :returns: solution x such that A.x=b
        """
        scipy_linear_solver = LinearSolver._check_linear_solver(linear_solver)

        # check the dimensions of b
        b_vec = LinearSolver._check_b(a_mat, b_vec)
        # solve the system
        if "tol" not in kwargs_lin:
            kwargs_lin["tol"] = 1e-8
        kwargs_lin["atol"] = kwargs_lin["tol"]

        if "maxiter" not in kwargs_lin:
            kwargs_lin["maxiter"] = 50 * len(b_vec)
        else:
            kwargs_lin["maxiter"] = min(kwargs_lin["maxiter"], 50 * len(b_vec))
        sol, info = scipy_linear_solver(
            A=a_mat, b=b_vec, outer_v=self.outer_v, **kwargs_lin
        )
        base_msg = "scipy linear solver algorithm stop info: "
        if info > 0:
            msg = "convergence to tolerance not achieved, number of iterations"
            total_msg = base_msg + msg
            LOGGER.warning(total_msg)
            total_msg = base_msg + "--- trying bicgstab method"
            LOGGER.warning(total_msg)

            sol, info = bicgstab(
                a_mat, b_vec, sol, maxiter=50 * len(b_vec), atol=kwargs_lin["atol"]
            )
            diff = a_mat.dot(sol) - b_vec.T
            res = np.sqrt(np.sum(diff))

            total_msg = "{} --- --- residual = {}".format(base_msg, res)
            LOGGER.warning(total_msg)
            total_msg = "{} --- --- info = {}".format(base_msg, info)
            LOGGER.warning(total_msg)

            if info < 0:
                total_msg = "{} --- trying cgs method".format(base_msg)
                LOGGER.warning(total_msg)

                sol, info = cgs(
                    a_mat, b_vec, sol, maxiter=50 * len(b_vec), atol=kwargs_lin["atol"]
                )
                diff = a_mat.dot(sol) - b_vec.T
                res = np.sqrt(np.sum(diff))

                total_msg = "{} --- --- residual = {}".format(base_msg, res)
                LOGGER.warning(total_msg)
                total_msg = "{} --- --- info = {}".format(base_msg, info)
                LOGGER.warning(total_msg)
        elif info < 0:
            msg = "illegal input or breakdown"
            total_msg = base_msg + msg
            LOGGER.error(total_msg)
        return np.atleast_2d(sol).T



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A k-means classification of the optimization history."""
from __future__ import division, unicode_literals

import logging
from typing import Optional, Union

from numpy import array
from numpy import int as np_int
from sklearn import cluster
from sklearn.preprocessing import StandardScaler

from gemseo.post.opt_post_processor import OptPostProcessor, OptPostProcessorOptionType
from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)


class KMeans(OptPostProcessor):
    """The **KMeans** post processing performs a k-means clustering on optimization
    history.

    The default number of clusters is 5 and can be modified in option.

    The k-means construction depends
    on the :code:`MiniBatchKMeans` class
    of the :code:`cluster` module of the
    `scikit-learn library <https://scikit-learn.org/stable/modules/generated/
    sklearn.cluster.MiniBatchKMeans.html>`_ .
    """

    def _run(
        self,
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Path]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_extension=None,  # type: Optional[str]
        **options  # type: OptPostProcessorOptionType
    ):  # type: (...) -> None
        self.__build_clusters(**options)

    def __build_clusters(
        self,
        n_clusters=5,  # type: int
    ):  # type: (...) -> None
        """Build the clusters.

        Args:
            n_clusters: The number of clusters.
        """
        x_history = self.database.get_x_history()
        x_vars = array(x_history)
        x_vars_sc = StandardScaler().fit_transform(x_vars)
        # estimate bandwidth for mean shift
        algorithm = cluster.MiniBatchKMeans(n_clusters=n_clusters)
        # predict cluster memberships
        algorithm.fit(x_vars_sc)
        y_pred = algorithm.labels_.astype(np_int)
        for x_vars, y_vars in zip(x_history, y_pred):
            self.database.store(x_vars, {"KM_cluster": int(y_vars)})
            self.out_data_dict[tuple(x_vars.real)] = y_vars



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The root mean squared error to measure the quality of a regression algorithm.

The :mod:`~gemseo.mlearning.qual_measure.mse_measure` module
implements the concept of root mean squared error measures
for machine learning algorithms.

This concept is implemented through the
:class:`.RMSEMeasure` class and
overloads the :meth:`!MSEMeasure.evaluate_*` methods.

The root mean squared error (RMSE) is defined by

.. math::

    \\operatorname{RMSE}(\\hat{y})=\\sqrt{\\frac{1}{n}\\sum_{i=1}^n(\\hat{y}_i-y_i)^2},

where
:math:`\\hat{y}` are the predictions and
:math:`y` are the data points.
"""
from __future__ import division, unicode_literals

from typing import List, Optional, Union

from numpy import ndarray

from gemseo.core.dataset import Dataset
from gemseo.mlearning.qual_measure.mse_measure import MSEMeasure
from gemseo.mlearning.regression.regression import MLRegressionAlgo


class RMSEMeasure(MSEMeasure):
    """The root mean Squared Error measure for machine learning."""

    def __init__(
        self,
        algo,  # type: MLRegressionAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for regression.
        """
        super(RMSEMeasure, self).__init__(algo)

    def evaluate_learn(
        self,
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        mse = super(RMSEMeasure, self).evaluate_learn(samples, multioutput)
        return mse ** 0.5

    def evaluate_test(
        self,
        test_data,  # type:Dataset
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        mse = super(RMSEMeasure, self).evaluate_test(test_data, samples, multioutput)
        return mse ** 0.5

    def evaluate_kfolds(
        self,
        n_folds=5,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        mse = super(RMSEMeasure, self).evaluate_kfolds(n_folds, samples, multioutput)
        return mse ** 0.5

    def evaluate_bootstrap(
        self,
        n_replicates=100,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        mse = super(RMSEMeasure, self).evaluate_bootstrap(
            n_replicates, samples, multioutput
        )
        return mse ** 0.5



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""This module contains the base class for the supervised machine learning algorithms.

Supervised machine learning is a task of learning relationships
between input and output variables based on an input-output dataset.
One usually distinguishes between two types of supervised machine learning algorithms,
based on the nature of the outputs.
For a continuous output variable,
a *regression* is performed,
while for a discrete output variable,
a *classification* is performed.

Given a set of input variables
:math:`x \\in \\mathbb{R}^{n_{\\text{samples}}\\times n_{\\text{inputs}}}` and
a set of output variables
:math:`y\\in \\mathbb{K}^{n_{\\text{samples}}\\times n_{\\text{outputs}}}`,
where :math:`n_{\\text{inputs}}` is the dimension of the input variable,
:math:`n_{\\text{outputs}}` is the dimension of the output variable,
:math:`n_{\\text{samples}}` is the number of training samples and
:math:`\\mathbb{K}` is either :math:`\\mathbb{R}` or :math:`\\mathbb{N}`
for regression and classification tasks respectively,
a supervised learning algorithm seeks to find a function
:math:`f: \\mathbb{R}^{n_{\\text{inputs}}} \\to
\\mathbb{K}^{n_{\\text{outputs}}}` such that :math:`y=f(x)`.

In addition,
we often want to impose some additional constraints on the function :math:`f`,
mainly to ensure that it has a generalization capacity beyond the training data,
i.e. it is able to correctly predict output values of new input values.
This is called regularization.
Assuming :math:`f` is parametrized by a set of parameters :math:`\\theta`,
and denoting :math:`f_\\theta` the parametrized function,
one typically seeks to minimize a function of the form

.. math::

    \\mu(y, f_\\theta(x)) + \\Omega(\\theta),

where :math:`\\mu` is a distance-like measure,
typically a mean squared error,
a cross entropy in the case of a regression,
or a probability to be maximized in the case of a classification,
and :math:`\\Omega` is a regularization term that limits the parameters
from over-fitting, typically some norm of its argument.

The :mod:`~gemseo.mlearning.core.supervised` module implements this concept
through the :class:`.MLSupervisedAlgo` class based on a :class:`.Dataset`.
"""
from __future__ import division, unicode_literals

from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from numpy import atleast_2d, ndarray

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, MLAlgo, MLAlgoParameterType
from gemseo.mlearning.core.ml_algo import SavedObjectType as MLAlgoSaveObjectType
from gemseo.mlearning.core.ml_algo import TransformerType
from gemseo.mlearning.transform.dimension_reduction.dimension_reduction import (
    DimensionReduction,
)
from gemseo.mlearning.transform.scaler.min_max_scaler import MinMaxScaler
from gemseo.utils.data_conversion import DataConversion

SavedObjectType = Union[MLAlgoSaveObjectType, Sequence[str], Dict[str, ndarray]]


class MLSupervisedAlgo(MLAlgo):
    """Supervised machine learning algorithm.

    Inheriting classes shall overload the :meth:`!MLSupervisedAlgo._fit` and
    :meth:`!MLSupervisedAlgo._predict` methods.

    Attributes:
        input_names (List[str]): The names of the input variables.
        output_names (List[str]): The names of the output variables.
        input_space_center (Dict[str,ndarray]): The center of the input space.
    """

    ABBR = "MLSupervisedAlgo"
    DEFAULT_TRANSFORMER = {Dataset.INPUT_GROUP: MinMaxScaler()}

    def __init__(
        self,
        data,  # type: Dataset
        transformer=DEFAULT_TRANSFORMER,  # type: TransformerType
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        **parameters  # type: MLAlgoParameterType
    ):  # type: (...) -> None
        """
        Args:
            input_names: The names of the input variables.
                If None, consider all input variables mentioned in the learning dataset.
            output_names: The names of the output variables.
                If None, consider all input variables mentioned in the learning dataset.
        """
        super(MLSupervisedAlgo, self).__init__(
            data, transformer=transformer, **parameters
        )
        self.input_names = input_names or data.get_names(data.INPUT_GROUP)
        self.output_names = output_names or data.get_names(data.OUTPUT_GROUP)
        self.input_space_center = None

    class DataFormatters(MLAlgo.DataFormatters):
        """Decorators for supervised algorithms."""

        @staticmethod
        def _array_to_dict(
            data_array,  # type: ndarray
            data_names,  # type: Sequence[str],
            data_sizes,  # type: Mapping[str,int]
        ):  # type: (...) -> Dict[str,ndarray]
            """Convert a NumPy data array into a data dictionary.

            Args:
                data_array: The data to be converted.
                data_names: The names of the variables.
                data_sizes: The sizes of the variables.

            Returns:
                The keys are the names of the variables
                and the values are their values.
            """
            current_position = 0
            array_dict = {}
            for name in data_names:
                array_dict[name] = data_array[
                    ..., current_position : current_position + data_sizes[name]
                ]
                current_position += data_sizes[name]
            return array_dict

        @classmethod
        def format_dict(
            cls,
            predict,  # type: Callable[[ndarray],ndarray]
        ):  # type: (...)-> Callable[[DataType],DataType]
            """Make an array-based function be called with a dictionary of NumPy arrays.

            Args:
                predict: The function to be called;
                    it takes a NumPy array in input and returns a NumPy array.

            Returns:
                A function making the function 'predict' work with
                either a NumPy data array
                or a dictionary of NumPy data arrays indexed by variables names.
                The evaluation will have the same type as the input data.
            """

            def wrapper(
                self,
                input_data,  # type: DataType
                *args,
                **kwargs
            ):  # type: (...) -> DataType
                """Evaluate 'predict' with either array or dictionary-based input data.

                Firstly,
                the pre-processing stage converts the input data to a NumPy data array,
                if these data are expressed as a dictionary of NumPy data arrays.

                Then,
                the processing evaluates the function 'predict'
                from this NumPy input data array.

                Lastly,
                the post-processing transforms the output data
                to a dictionary of output NumPy data array
                if the input data were passed as a dictionary of NumPy data arrays.

                Args:
                    input_data: The input data.
                    *args: The positional arguments of the function 'predict'.
                    **kwargs: The keyword arguments of the function 'predict'.

                Returns:
                    The output data with the same type as the input one.
                """
                as_dict = isinstance(input_data, dict)
                if as_dict:
                    input_data = DataConversion.dict_to_array(
                        input_data, self.input_names
                    )
                output_data = predict(self, input_data, *args, **kwargs)
                if as_dict:
                    varsizes = self.learning_set.sizes
                    output_data = cls._array_to_dict(
                        output_data, self.output_names, varsizes
                    )
                return output_data

            return wrapper

        @classmethod
        def format_samples(
            cls,
            predict,  # type: Callable[[ndarray],ndarray]
        ):  # type: (...)-> Callable[[ndarray],ndarray]
            """Make a 2D NumPy array-based function work with 1D NumPy array.

            Args:
                predict: The function to be called;
                    it takes a 2D NumPy array in input
                    and returns a 2D NumPy array.
                    The first dimension represents the samples
                    while the second one represents the components of the variables.

            Returns:
                A function making the function 'predict' work with
                either a 1D NumPy array or a 2D NumPy array.
                The evaluation will have the same dimension as the input data.
            """

            def wrapper(
                self,
                input_data,  # type: DataType
                *args,
                **kwargs
            ):  # type: (...) -> DataType
                """Evaluate 'predict' with either a 1D or 2D NumPy data array.

                Firstly,
                the pre-processing stage converts the input data
                to a 2D NumPy data array.

                Then,
                the processing evaluates the function 'predict'
                from this 2D NumPy data array.

                Lastly,
                the post-processing converts the output data to a 1D NumPy data array
                if the dimension of the input data is equal to 1.

                Args:
                    input_data: The input data.
                    *args: The positional arguments of the function 'predict'.
                    **kwargs: The keyword arguments of the function 'predict'.

                Returns:
                    The output data with the same dimension as the input one.
                """
                single_sample = input_data.ndim == 1
                input_data = atleast_2d(input_data)
                output_data = predict(self, input_data, *args, **kwargs)
                if single_sample:
                    output_data = output_data[0]
                return output_data

            return wrapper

        @classmethod
        def format_transform(
            cls,
            transform_inputs=True,  # type: bool
            transform_outputs=True,  # type:bool
        ):  # type: (...)-> Callable[[ndarray],ndarray]
            """Force a function to transform its input and/or output variables.

            Args:
                transform_inputs: If True,
                    apply the transformers to the input variables.
                transform_outputs: If True,
                    apply the transformers to the output variables.

            Returns:
                A function evaluating a function of interest,
                after transforming its input data
                and/or before transforming its output data.
            """

            def format_transform_(
                predict,  # type: Callable[[ndarray],ndarray]
            ):  # type: (...)-> Callable[[ndarray],ndarray]
                """Apply transformation to inputs and inverse transformation to outputs.

                Args:
                    predict: The function of interest to be called.

                Returns:
                    A function evaluating the function 'predict',
                    after transforming its input data
                    and/or before transforming its output data.
                """

                def wrapper(
                    self,
                    input_data,  # type: DataType
                    *args,
                    **kwargs
                ):  # type: (...) -> DataType
                    """Evaluate 'predict' after or before data transformation.

                    Firstly,
                    the pre-processing stage transforms the input data if required.

                    Then,
                    the processing evaluates the function 'predict'.

                    Lastly,
                    the post-processing stage transforms the output data if required.

                    Args:
                        input_data: The input data.
                        *args: The positional arguments of the function.
                        **kwargs: The keyword arguments of the function.

                    Returns:
                        Either the raw output data of 'predict'
                        or a transformed version according to the requirements.
                    """
                    inputs = self.learning_set.INPUT_GROUP
                    if transform_inputs and inputs in self.transformer:
                        input_data = self.transformer[inputs].transform(input_data)
                    output_data = predict(self, input_data, *args, **kwargs)
                    outputs = self.learning_set.OUTPUT_GROUP
                    if transform_outputs and outputs in self.transformer:
                        output_data = self.transformer[outputs].inverse_transform(
                            output_data
                        )
                    return output_data

                return wrapper

            return format_transform_

        @classmethod
        def format_input_output(
            cls,
            predict,  # type: Callable[[ndarray],ndarray]
        ):  # type: (...)-> Callable[[DataType],DataType]
            """Make a function robust to type, array shape and data transformation.

            Args:
                predict: The function of interest to be called.

            Returns:
                A function calling the function of interest 'predict',
                while guaranteeing consistency in terms of data type and array shape,
                and applying input and/or output data transformation if required.
            """

            @cls.format_dict
            @cls.format_samples
            @cls.format_transform()
            def wrapper(self, input_data, *args, **kwargs):
                return predict(self, input_data, *args, **kwargs)

            return wrapper

    def learn(
        self,
        samples=None,  # type: Optional[List[int]]
    ):  # type: (...) -> None
        """
        Raises:
            NotImplementedError: If an output transformer modifies
                both the input and the output variables, e.g. :class:`PLS`.
        """
        input_grp = self.learning_set.INPUT_GROUP
        output_grp = self.learning_set.OUTPUT_GROUP
        input_data = self.learning_set.get_data_by_names(self.input_names, False)
        output_data = self.learning_set.get_data_by_names(self.output_names, False)

        if samples is not None:
            input_data = input_data[samples]
            output_data = output_data[samples]

        self.input_space_center = DataConversion.array_to_dict(
            input_data.mean(0), self.input_names, self.learning_set.sizes
        )

        if input_grp in self.transformer:
            transformer = self.transformer[input_grp]
            if transformer.CROSSED:
                input_data = transformer.fit_transform(input_data, output_data)
            else:
                input_data = transformer.fit_transform(input_data)

        if output_grp in self.transformer:
            transformer = self.transformer[output_grp]
            if self.transformer[output_grp].CROSSED:
                raise NotImplementedError(
                    "The transformer of type {} cannot be applied to the outputs "
                    "to build a supervised machine learning algorithm".format(
                        self.transformer[output_grp].__class__.__name__
                    )
                )
            else:
                output_data = transformer.fit_transform(output_data)

        self._fit(input_data, output_data)
        self._trained = True

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Fit input-output relationship from the learning data.

        Args:
            input_data: The input data with the shape (n_samples, n_inputs).
            output_data: The output data with shape (n_samples, n_outputs).
        """
        raise NotImplementedError

    @DataFormatters.format_input_output
    def predict(
        self,
        input_data,  # type: DataType
    ):  # type: (...) -> DataType
        """Predict output data from input data.

        The user can specify these input data either as a NumPy array,
        e.g. :code:`array([1., 2., 3.])`
        or as a dictionary,
        e.g.  :code:`{'a': array([1.]), 'b': array([2., 3.])}`.

        If the numpy arrays are of dimension 2,
        their i-th rows represent the input data of the i-th sample;
        while if the numpy arrays are of dimension 1,
        there is a single sample.

        The type of the output data and the dimension of the output arrays
        will be consistent
        with the type of the input data and the size of the input arrays.

        Args:
            input_data: The input data.

        Returns:
            The predicted output data.
        """
        return self._predict(input_data)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Predict output data from input data.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            output_data: The output data with shape (n_samples, n_outputs).
        """
        raise NotImplementedError

    def _get_raw_shapes(self):  # type: (...) -> Tuple[int,int]
        """Get the raw input and output shapes.

        The raw shapes are the shapes of the input and output variables
        after applying the transformers.

        Returns:
            The raw input and output shapes.
        """
        reduce_inputs = Dataset.INPUT_GROUP in self.transformer and isinstance(
            self.transformer[Dataset.INPUT_GROUP], DimensionReduction
        )
        if reduce_inputs:
            input_shape = self.transformer[Dataset.INPUT_GROUP].n_components
        else:
            input_shape = self.input_shape

        reduce_outputs = Dataset.OUTPUT_GROUP in self.transformer and isinstance(
            self.transformer[Dataset.OUTPUT_GROUP], DimensionReduction
        )
        if reduce_outputs:
            output_shape = self.transformer[Dataset.OUTPUT_GROUP].n_components
        else:
            output_shape = self.output_shape

        return input_shape, output_shape

    @property
    def input_shape(self):  # type: (...) -> int
        """The dimension of the input variables before applying the transformers."""
        sizes = [self.learning_set.sizes[name] for name in self.input_names]
        return sum(sizes)

    @property
    def output_shape(self):  # type: (...) -> int
        """The dimension of the output variables before applying the transformers."""
        sizes = [self.learning_set.sizes[name] for name in self.output_names]
        return sum(sizes)

    @property
    def input_data(self):  # type: (...) -> ndarray
        """The input data matrix."""
        in_names = self.input_names
        inputs = self.learning_set.get_data_by_names(in_names, False)
        return inputs

    @property
    def output_data(self):  # type: (...) -> ndarray
        """The output data matrix."""
        out_names = self.output_names
        outputs = self.learning_set.get_data_by_names(out_names, False)
        return outputs

    def _get_objects_to_save(self):  # type: (...) -> Dict[str,SavedObjectType]
        objects = super(MLSupervisedAlgo, self)._get_objects_to_save()
        objects["input_names"] = self.input_names
        objects["output_names"] = self.output_names
        objects["input_space_center"] = self.input_space_center
        return objects



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A Gauss Seidel algorithm for solving MDAs
*****************************************
"""
from __future__ import division, unicode_literals

from gemseo.core.chain import MDOChain
from gemseo.core.discipline import MDODiscipline
from gemseo.mda.mda import MDA


class MDAGaussSeidel(MDA):
    """Perform a MDA analysis using a Gauss-Seidel algorithm, an iterative technique to
    solve the linear system:

    .. math::

       Ax = b

    by decomposing the matrix :math:`A`
    into the sum of a lower triangular matrix :math:`L_*`
    and a strictly upper triangular matrix :math:`U`.

    The new iterate is given by:

    .. math::

       x_{k+1} = L_*^{-1}(b-Ux_k)
    """

    def __init__(
        self,
        disciplines,
        name=None,
        max_mda_iter=10,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        tolerance=1e-6,
        linear_solver_tolerance=1e-12,
        warm_start=False,
        use_lu_fact=False,
        over_relax_factor=1.0,
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param name: the name of the chain
        :type name: str
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        :param over_relax_factor: relaxation coefficient, used to make the
            method more robust, if 0<over_relax_factor<1
            or faster if 1<over_relax_factor<=2.
            If over_relax_factor =1., it is deactivated
        :type over_relax_factor: float
        """
        self.chain = MDOChain(disciplines, grammar_type=grammar_type)
        super(MDAGaussSeidel, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            grammar_type=grammar_type,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
            use_lu_fact=use_lu_fact,
        )
        assert over_relax_factor > 0.0
        assert over_relax_factor <= 2.0
        self.over_relax_factor = over_relax_factor
        self._initialize_grammars()
        self._set_default_inputs()
        self._compute_input_couplings()

    def _initialize_grammars(self):
        """Defines all inputs and outputs of the chain."""
        # self.chain.initialize_grammars()
        self.input_grammar.update_from(self.chain.input_grammar)
        self.output_grammar.update_from(self.chain.output_grammar)

    def _run(self):
        """Runs the disciplines in a sequential way until the difference between outputs
        is under tolerance.

        :returns: the local data
        """
        if self.warm_start:
            self._couplings_warm_start()
        current_couplings = 0.0

        relax = self.over_relax_factor
        use_relax = relax != 1.0
        # store initial residual
        current_iter = 0
        while not self._termination(current_iter) or current_iter == 0:
            for discipline in self.disciplines:
                discipline.execute(self.local_data)
                outs = discipline.get_output_data()
                if use_relax:
                    # First time this output is computed, update directly local data
                    self.local_data.update(
                        {k: v for k, v in outs.items() if k not in self.local_data}
                    )
                    # The couplings already exist in the local data,
                    # so the over relaxation can be applied
                    self.local_data.update(
                        {
                            k: relax * v + (1.0 - relax) * self.local_data[k]
                            for k, v in outs.items()
                            if k in self.local_data
                        }
                    )
                else:
                    self.local_data.update(outs)

            new_couplings = self._current_strong_couplings()
            self._compute_residual(
                current_couplings, new_couplings, current_iter, first=current_iter == 0
            )
            # store current residuals
            current_iter += 1
            current_couplings = new_couplings

        for discipline in self.disciplines:  # Update all outputs without relax
            self.local_data.update(discipline.get_output_data())



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""JSON schema handler."""

from gemseo.utils.py23_compat import PY2

if PY2:
    from collections import ItemsView, KeysView, Mapping, ValuesView
else:
    from collections.abc import ItemsView, KeysView, Mapping, ValuesView

from genson import SchemaBuilder
from genson.schema.strategies import Object


class _MergeRequiredStrategy(Object):
    """A genson Object with a modified required attribute handling.

    Genson Object does not merge the required attribute on purpose.
    See :ref:`https://github.com/wolverdude/GenSON#genson`.
    This class will merge the required attributes.
    """

    # Do not merge the name and id properties.
    KEYWORDS = Object.KEYWORDS + ("name", "id")

    def add_schema(self, schema):
        """Add a schema and merge the required attribute.

        Args:
            schema: A schema to be added.
        """
        if "required" not in schema or self._required is None:
            super(_MergeRequiredStrategy, self).add_schema(schema)
        else:
            required = set(self._required)
            super(_MergeRequiredStrategy, self).add_schema(schema)
            self._required = required | set(schema["required"])

    def add_object(self, obj):
        """Add an object and merge the required attribute.

        Args:
            obj: An object to be added.
        """
        if self._required is None:
            super(_MergeRequiredStrategy, self).add_object(obj)
        else:
            required = set(self._required)
            super(_MergeRequiredStrategy, self).add_object(obj)
            self._required = required | set(obj.keys())


class MutableMappingSchemaBuilder(SchemaBuilder):
    """A mutable genson SchemaBuilder with a dictionary-like interface.

    The mutability is provided by the ability to delete a property.
    """

    EXTRA_STRATEGIES = (_MergeRequiredStrategy,)

    def get(self, key, default=None):
        """Implement the standard mapping getter.

        Args:
            key: A key whose mapped value shall be returned.
            default: The default value returned if the key is missing.

        Returns:
            The value mapped to key if it exists, default otherwise.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def keys(self):
        """Return the keys.

        Returns:
            The keys.
        """
        return KeysView(self)

    def items(self):
        """Return the pairs of key and mapped value.

        Returns:
            The pairs of key and mapped value.
        """
        return ItemsView(self)

    def values(self):
        """Return the values.

        Returns:
            The values.
        """
        return ValuesView(self)

    def __eq__(self, other):
        if not isinstance(other, Mapping):
            return NotImplemented
        return dict(self.items()) == dict(other.items())

    __reversed__ = None

    @property
    def _properties(self):
        """Return the properties.

        Returns:
            The existing properties, otherwise an empty dictionary.
        """
        try:
            return self._root_node._active_strategies[0]._properties
        except AttributeError:
            return {}

    def __getitem__(self, key):
        return self._properties[key]

    def __iter__(self):
        return iter(self._properties)

    def __len__(self):
        return len(self._properties)

    def __delitem__(self, key):
        del self._properties[key]
        try:
            self._root_node._active_strategies[0]._required.remove(key)
        except IndexError:
            pass

    # For backward compatibility
    # TODO: add deprecation warning
    to_dict = SchemaBuilder.to_schema



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Services for handling file paths."""
import re
from collections import namedtuple
from enum import Enum
from re import findall
from typing import Optional, Union

from gemseo.utils.py23_compat import Path
from gemseo.utils.string_tools import MultiLineString

FileDefinition = namedtuple("FileDefinition", ["name", "extension"])


class FileType(Enum):
    """The type of a file, defined by its default name and format."""

    FIGURE = FileDefinition("figure", "png")
    SCHEMA = FileDefinition("schema", "json")
    TEXT = FileDefinition("document", "txt")
    WEBPAGE = FileDefinition("page", "html")


class FilePathManager(object):
    """A factory of file paths for a given type of file and with default settings."""

    def __init__(
        self,
        file_type,  # type: FileType
        default_name=None,  # type: str
        default_directory=None,  # type:Optional[Path]
        default_extension=None,  # type: Optional[str]
    ):  # type: (...) -> None
        """
        Args:
            file_type: The type of file,
                defined by its default file name and format;
                select a file type by iterating over ``FileType``.
            default_name: The default file name.
                If None, use the default file name related to the given type of file.
            default_directory: The default directory path.
                If None, use the current working directory.
            default_extension: The default extension.
                If None, use the default extension related to the given type of file.
        """
        if default_name is None:
            self.__default_name = file_type.value.name
        else:
            self.__default_name = default_name

        self.__default_directory = Path(default_directory or Path.cwd())

        if default_extension is None:
            self.__default_extension = file_type.value.extension
        else:
            self.__default_extension = default_extension

        self.__file_type = file_type.name

    def __str__(self):  # type: (...) -> str
        string = MultiLineString()
        string.add(self.__class__.__name__)
        string.indent()
        string.add("File type: {}", self.__file_type)
        string.add("Default file name: {}", self.__default_name)
        string.add("Default file extension: {}", self.__default_extension)
        string.add("Default directory: {}", self.__default_directory)
        return str(string)

    def create_file_path(
        self,
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_extension=None,  # type: Optional[str]
    ):  # type: (...) -> Path
        """Make a file path from a directory path, a file name and a file extension.

        Args:
            file_path: The path of the file to be returned.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory.
                If None, use the default directory path.
            file_name: The file name to be used.
                If None, use the default file name.
            file_extension: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use the default file extension.

        Returns:
            The file path.
        """
        suffix = ".{}".format(file_extension or self.__default_extension)

        if file_path is not None:
            file_path = Path(file_path)
            if not file_path.suffix:
                file_path = file_path.with_suffix(suffix)
            return file_path

        file_name = file_name or self.__default_name

        if directory_path is None:
            directory_path = self.__default_directory
        else:
            directory_path = Path(directory_path)

        return (directory_path / file_name).with_suffix(suffix)

    @staticmethod
    def to_snake_case(
        message,  # type: str
    ):  # type: (...) -> str
        """Snake case a string.

        That means:

        1. Split the message.
        2. Lowercase the resulting elements.

        ``-`` and `` `` are replaced with ``_``.

        Args:
            message: The message to be snake-cased.

        Returns:
            The snake-cased message.
        """
        message = message.replace("-", "_").replace(" ", "_")
        message = "_".join(
            [elem.lower() for elem in findall("[A-Z][^A-Z]*", message)] or [message]
        )
        return re.sub("_+", "_", message)

    @classmethod
    def add_suffix(
        cls,
        file_path,  # type: Path
        suffix,  # type: str
    ):  # type: (...) -> Path
        """Add a suffix to an existing file path between the filename and the extension.

        E.g. `directory/filename_suffix.pdf`.

        Args:
            file_path: The file path to be suffixed.
            suffix: The suffix to be added to the file path.

        Returns:
            The directory path, the file name and the file extension
            obtained from the file path.
        """
        extension = file_path.suffix
        file_name = "{}_{}".format(str(file_path.stem), suffix)
        directory_path = file_path.parent
        return (directory_path / file_name).with_suffix(extension)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Damien Guenot
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
The Rastrigin analytic problem
******************************
"""
from __future__ import division, unicode_literals

import logging
from cmath import cos, pi, sin

from numpy import array, ones, zeros

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.function import MDOFunction

LOGGER = logging.getLogger(__name__)


class Rastrigin(OptimizationProblem):

    r"""**Rastrigin** :class:`.OptimizationProblem`
    uses the Rastrigin objective function
    with the :class:`.DesignSpace` :math:`[-0.1,0.1]^2`

    From http://en.wikipedia.org/wiki/Rastrigin_function:

        the Rastrigin function is a non-convex function used as a
        performance test problem for optimization algorithms.
        It is a typical example of non-linear multimodal function.
        It was first proposed by [Rastrigin] as a 2-dimensional
        function and has been generalized by [MuhlenbeinEtAl].
        Finding the minimum of this function is a fairly difficult
        problem due to its large search space and its large
        number of local minima.
        It has a global minimum at :math:`x=0` where :math:`f(x)=0`.
        It can be extended to :math:`n>2` dimensions:

        .. math::

           f(x) = 10n + \sum_{i=1}^n [x_i^2 - 10\cos(2\pi x_i)]

        [Rastrigin] Rastrigin, L. A. "Systems of extremal control."
        Mir, Moscow (1974).

        [MuhlenbeinEtAl] H. Mühlenbein, D. Schomisch and J. Born.
        "The Parallel Genetic Algorithm as Function Optimizer ".
        Parallel Computing, 17, pages 619–632, 1991.
    """

    def __init__(self):
        """
        The constructor initializes the Rastrigin
        :class:`.OptimizationProblem`
        by defining the :class:`.DesignSpace`
        and the objective function.
        """
        design_space = DesignSpace()
        design_space.add_variable("x", 2, l_b=-0.1, u_b=0.1)
        design_space.set_current_x(0.01 * ones(2))
        super(Rastrigin, self).__init__(design_space)
        expr = "20 + sum(x[i]**2 - 10*cos(2pi*x[i]))"
        self.objective = MDOFunction(
            self.rastrigin,
            name="Rastrigin",
            f_type="obj",
            jac=self.rastrigin_jac,
            expr=expr,
            args=["x"],
        )

    @staticmethod
    def rastrigin(x_dv):
        """This function computes the order n=2 Rastrigin function.

        :param x_dv: design variable vector of size 2
        :returns: result of Rastrigin function evaluation
        """
        a_c = 10.0
        func = (
            a_c * 2.0
            + (x_dv[0] ** 2 - a_c * cos(2 * pi * x_dv[0]))
            + (x_dv[1] ** 2 - a_c * cos(2 * pi * x_dv[1]))
        )
        return func.real

    @staticmethod
    def get_solution():
        """Return theoretical optimal value of Rastrigin function.

        :returns: design variables values of optimized values,
            function value at optimum
        :rtype: numpy array
        """
        x_opt = zeros(2)
        f_opt = 0.0
        return x_opt, f_opt

    @staticmethod
    def rastrigin_jac(x_dv):
        """This function computes the analytical gradient of 2nd order Rastrigin
        function.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: analytical gradient vector of Rastrigin function
        :rtype: numpy array
        """
        a_c = 10.0
        analytic_grad = array(
            [
                2 * x_dv[0] + 2 * pi * a_c * sin(2 * pi * x_dv[0]),
                2 * x_dv[1] + 2 * pi * a_c * sin(2 * pi * x_dv[1]),
            ]
        )
        return analytic_grad.real



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Helper functions for hdf5 data."""

from typing import Union

import h5py


def get_hdf5_group(
    h5py_data,  # type: Union[h5py.File, h5py.Group]
    name,  # type: str
):  # type: (...) -> None
    """Return a group from a h5py data handle.

    This function shall be used to show a better error message to the end user.

    Args:
        h5py_data: The hdf5 data handle.
        name: The name of the group.

    Returns:
        The contents of the group.

    Raises:
        KeyError if the group does not exist.
    """
    try:
        return h5py_data[name]
    except KeyError as err:
        raise KeyError(
            "In hdf5 file {}: no such group {}".format(h5py_data.file, err.args[0])
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Computation of tolerance intervals from a data-fitted exponential distribution."""

import openturns as ot

from gemseo.uncertainty.statistics.tolerance_interval.distribution import (
    ToleranceInterval,
)


class ExponentialToleranceInterval(ToleranceInterval):
    """Computation of tolerance intervals from a data-fitted exponential distribution.

    The formulae come from the R library *tolerance* [1]_.

    .. [1] Derek S. Young,
           *tolerance: An R Package for Estimating Tolerance Intervals*,
           Journal of Statistical Software, 36(5), 2010
    """

    def __init__(
        self,
        size,  # type: int
        rate,  # type: float
        location,  # type: float
    ):  # type:(...) -> None
        # noqa: D205 D212 D415
        """
        Args:
            rate: The estimation of the rate of the exponential distribution.
            location: The estimation of the location of the exponential distribution.
        """
        super(ExponentialToleranceInterval, self).__init__(size)
        self.__rate = rate
        self.__location = location

    def _compute_lower_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        k_1 = 1 - (coverage ** size / alpha) ** (1.0 / (size - 1))
        return self.__location + k_1 / self.__rate

    def _compute_upper_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        chi2_num = ot.ChiSquare(2).computeQuantile(coverage)[0]
        chi2_den = ot.ChiSquare(2 * size - 2).computeQuantile(coverage)[0]
        k_2 = size * chi2_num / chi2_den
        return self.__location + k_2 / self.__rate



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A factory to create instances of :class:`.DatasetPlot`.

The module :mod:`~gemseo.post.dataset.factory` contains the :class:`DatasetPlotFactory`
class which is a factory to instantiate a :class:`.DatasetPlot` from its class name. The
class can be internal to |g| or located in an external module whose path is provided to
the constructor. It also provides a list of available cache types and allows you to test
if a cache type is available.
"""
from __future__ import division, unicode_literals

import logging
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from gemseo.core.dataset import Dataset

from gemseo.core.factory import Factory
from gemseo.post.dataset.dataset_plot import DatasetPlot

LOGGER = logging.getLogger(__name__)


class DatasetPlotFactory(object):
    """This factory instantiates a :class:`.DatasetPlot` from its class name."""

    def __init__(self):  # type: (...) -> None
        self.factory = Factory(DatasetPlot, ("gemseo.post.dataset",))

    def create(
        self,
        plot_name,  # type: str
        dataset,  # type: Dataset
        **options
    ):  # type: (...) -> DatasetPlot
        """Create a plot for dataset.

        Args:
            plot_name: The name of a plot method for dataset (its class name).
            dataset: The dataset to visualize.
            options: The additional options specific to this plot method.

        Returns:
            A plot method built from the provided dataset.
        """
        return self.factory.create(plot_name, dataset=dataset, **options)

    @property
    def plots(self):  # type: (...) -> List[str]
        """The available plot methods for dataset."""
        return self.factory.classes

    def is_available(
        self,
        plot_name,  # type: str
    ):  # type: (...) -> bool
        """Check the availability of a plot for dataset.

        Args:
            plot_name: The name of a plot method for dataset (its class name).

        Returns:
            True if the plot method is available.
        """
        return self.factory.is_available(plot_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Damien Guenot
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
The Rosenbrock analytic problem
*******************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import array, atleast_2d, ones, zeros
from scipy.optimize import rosen, rosen_der

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.discipline import MDODiscipline
from gemseo.core.function import MDOFunction

LOGGER = logging.getLogger(__name__)


class Rosenbrock(OptimizationProblem):
    r"""**Rosenbrock** :class:`.OptimizationProblem`
    uses the Rosenbrock objective function

    .. math::

       f(x) = \sum_{i=2}^{n_x} 100(x_{i} - x_{i-1}^2)^2 + (1 - x_{i-1})^2

    with the default :class:`.DesignSpace`
    :math:`[-0.2,0.2]^{n_x}`.
    """

    def __init__(self, n_x=2, l_b=-2.0, u_b=2.0, scalar_var=False, initial_guess=None):
        """
        The constructor initializes the Rosenbrock
        :class:`.OptimizationProblem`
        by defining the :class:`.DesignSpace` and
        the objective function.

        :param n_x: problem dimension
        :type n_x: int
        :param l_b: lower bound (common value to all variables)
        :type l_b: float
        :param u_b: upper bound (common value to all variables)
        :type u_b: float
        :param scalar_var: if True the design space will
            contain only scalar variables
            (as many as the problem dimension); if False
            the design space will contain a
            single multidimensional variable (whose size
            equals the problem dimension)
        :type scalar_var: bool
        :param initial_guess: initial guess for optimal solution
        :type initial_guess: numpy array
        """
        design_space = DesignSpace()
        if scalar_var:
            args = ["x" + str(i) for i in range(1, n_x + 1)]
            for arg in args:
                design_space.add_variable(arg, l_b=l_b, u_b=u_b)
        else:
            args = ["x"]
            design_space.add_variable("x", size=n_x, l_b=l_b, u_b=u_b)
        if initial_guess is None:
            design_space.set_current_x(zeros(n_x))
        else:
            design_space.set_current_x(initial_guess)

        super(Rosenbrock, self).__init__(design_space)
        expr = "sum(100*(x[1:] - x[:-1]**2)**2 + (1 - x[:-1])**2"
        self.objective = MDOFunction(
            rosen,
            name="rosen",
            f_type=MDOFunction.TYPE_OBJ,
            jac=rosen_der,
            expr=expr,
            args=args,
        )

    def get_solution(self):
        """Return the theoretical optimal value.

        :returns: design variables values of optimized values,
            function value at optimum
        :rtype: numpy array
        """
        x_opt = ones(self.design_space.dimension)
        f_opt = 0.0
        return x_opt, f_opt


class RosenMF(MDODiscipline):
    r"""**RosenMF**, a multi-fidelity Rosenbrock
    :class:`.MDODiscipline`,
    returns the value:

    .. math::

       \mathrm{fidelity} * \mathrm{Rosenbrock}(x)

    where both :math:`\mathrm{fidelity}` and :math:`x`
    are provided as input data.
    """

    def __init__(self, dimension=2):
        """The constructor defines the default inputs of the :class:`.MDODiscipline`,
        namely the default design parameter values and the fidelity.

        :param dimension: problem dimension
        :type dimension: int
        """
        super(RosenMF, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = {"x": zeros(dimension), "fidelity": array([1.0])}

    def _run(self):
        fidelity = self.local_data["fidelity"]
        x_val = self.local_data["x"]
        self.local_data["rosen"] = fidelity * rosen(x_val)

    def _compute_jacobian(self, inputs=None, outputs=None):
        x_val = self.local_data["x"]
        fidelity = self.local_data["fidelity"]
        self.jac = {
            "rosen": {
                "x": atleast_2d(fidelity * rosen_der(x_val)),
                "fidelity": atleast_2d(rosen(x_val)),
            }
        }



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The polynomial chaos expansion algorithm for regression.

The polynomial chaos expansion (PCE) model expresses the model output
as a weighted sum of polynomial functions which are orthonormal
in the stochastic input space spanned by the random input variables:

.. math::

    Y = w_0 + w_1\phi_1(X) + w_2\phi_2(X) + ... + w_K\phi_K(X)

where :math:`\phi_i(x)=\psi_{\tau_1(i),1}(x_1)\times\ldots\times
\psi_{\tau_d(i),d}(x_d)`.

Enumerating strategy
--------------------

The choice of the function :math:`\tau=(\tau_1,\ldots,\tau_d)` is an
enumerating strategy and :math:`\tau_j(i)` is the polynomial degree of
:math:`\psi_{\tau_j(i),j}`.

Distributions
-------------

PCE are stochastic models whose inputs are random variables
and are often used to deal with uncertainty quantification problems.

If :math:`X_j` is a Gaussian random variable,
:math:`(\psi_{ij})_{i\geq 0}` is the Legendre basis.
If :math:`X_j` is an uniform random variable,
:math:`(\psi_{ij})_{i\geq 0}` is the Hermite basis.

When the problem is deterministic,
we can still use PCE under the assumptions that
the random variables are independent uniform random variables.
Then,
the orthonormal basis function is the Hermite basis.

Degree
------

The degree :math:`P` of a PCE is defined
in such a way that :math:`\text{degree}(\phi_i)=\sum_{j=1}^d\tau_j(i)\leq P`.

Estimation
----------

The coefficients :math:`(w_1, w_2, ..., w_K)` and the intercept :math:`w_0`
are estimated either by least squares regression,
sparse least squares regression or quadrature.

Dependence
----------
The PCE model relies on the FunctionalChaosAlgorithm class
of the `openturns library <https://openturns.github.io/openturns/latest/user_manual/
response_surface/_generated/openturns.FunctionalChaosAlgorithm.html>`_.
"""
from __future__ import division, unicode_literals

import logging
from typing import Dict, Iterable, Mapping, Optional, Tuple, Union

import openturns
from numpy import all as np_all
from numpy import array, concatenate, isin, ndarray, zeros, zeros_like
from openturns import (
    LARS,
    AdaptiveStieltjesAlgorithm,
    CleaningStrategy,
    ComposedDistribution,
    CorrectedLeaveOneOut,
    FixedStrategy,
    Function,
    FunctionalChaosAlgorithm,
    FunctionalChaosSobolIndices,
    GaussProductExperiment,
    HyperbolicAnisotropicEnumerateFunction,
    IntegrationStrategy,
    LeastSquaresMetaModelSelectionFactory,
    LeastSquaresStrategy,
    MarginalTransformationEvaluation,
    OrthogonalBasis,
    OrthogonalProductPolynomialFactory,
    Point,
    StandardDistributionPolynomialFactory,
)

from gemseo.algos.parameter_space import ParameterSpace
from gemseo.core.dataset import Dataset
from gemseo.core.discipline import MDODiscipline
from gemseo.mlearning.core.ml_algo import TransformerType
from gemseo.mlearning.regression.regression import MLRegressionAlgo

LOGGER = logging.getLogger(__name__)


class PCERegression(MLRegressionAlgo):
    """Polynomial chaos expansion."""

    LIBRARY = "openturns"
    ABBR = "PCE"
    LS_STRATEGY = "LS"
    QUAD_STRATEGY = "Quad"
    SPARSE_STRATEGY = "SparseLS"
    AVAILABLE_STRATEGIES = [LS_STRATEGY, QUAD_STRATEGY, SPARSE_STRATEGY]

    def __init__(
        self,
        data,  # type: Dataset
        probability_space,  # type: ParameterSpace
        discipline=None,  # type: Optional[MDODiscipline]
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        strategy=LS_STRATEGY,  # type: str
        degree=2,  # type: int
        n_quad=None,  # type:Optional[int]
        stieltjes=True,  # type: bool
        sparse_param=None,  # type: Optional[Mapping[str,Union[int,float]]]
    ):  # type: (...) -> None
        """
        Args:
            probability_space: The probability space
                defining the probability distributions of the model inputs.
            discipline: The discipline to evaluate with the quadrature strategy
                if the learning set does not have output data.
                If None, use the output data from the learning set.
            strategy: The strategy to compute the parameters of the PCE,
                either 'LS' for *least-square*, 'Quad' for *quadrature*
                or 'SparseLS' for *sparse least-square*.
            degree: The polynomial degree of the PCE.
            n_quad: The total number of quadrature points
                used by the quadrature strategy
                to compute the marginal number of points by input dimension.
                If None, this degree will be set equal
                to the polynomial degree of the PCE plus one.
            stieltjes: Use the Stieltjes method.
            sparse_param: The parameters for the Sparse Cleaning Truncation
                Strategy and/or hyperbolic truncation of the initial basis:

                - **max_considered_terms** (int) -- The maximum considered terms
                  (default: 120),
                - **most_significant** (int) -- The most Significant number to retain
                  (default: 30),
                - **significance_factor** (float) -- Significance Factor
                  (default: 1e-3),
                - **hyper_factor** (float) -- The factor for the hyperbolic truncation
                  strategy (default: 1.0).

                If None, use default values.

        Raises:
            ValueError: Either if the variables of the probability space
                and the input variables of the dataset are different,
                if transformers are specified for the inputs,
                or if the strategy to compute the parameters of the PCE is unknown.
        """
        prob_space = probability_space
        super(PCERegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            probability_space=prob_space,
            strategy=strategy,
            degree=degree,
            n_quad=n_quad,
            stieltjes=stieltjes,
            sparse_param=sparse_param,
        )
        self._prob_space = prob_space
        self._discipline = discipline
        try:
            if data:
                u_names = set(prob_space.variables_names)
                assert set(self.input_names) == u_names
            else:
                self.input_names = prob_space.variables_names
        except Exception:
            raise ValueError(
                "Data inputs names are %s "
                "while probability distributions are defined "
                "%s." % (self.input_names, list(prob_space.distributions.keys()))
            )
        forbidden_names = set(self.input_names).union({Dataset.INPUT_GROUP})
        if set(list(self.transformer.keys())).intersection(forbidden_names):
            raise ValueError("PCERegression does not support input transformers.")

        self._distributions = prob_space.distributions
        self._sparse_param = sparse_param or {}
        self._input_dim = sum(
            [dist.dimension for _, dist in self._distributions.items()]
        )
        self._strategy = strategy
        if self._strategy not in self.AVAILABLE_STRATEGIES:
            strategies = " ".join(self.AVAILABLE_STRATEGIES)
            raise ValueError(
                "{} is a wrong strategy."
                " Available ones are: {}.".format(self._strategy, strategies)
            )
        self._stieltjes = stieltjes
        self._degree = degree
        self._n_quad = n_quad
        self._ot_distributions = [
            self._distributions[name].distribution for name in self.input_names
        ]
        self._dist = ComposedDistribution(self._ot_distributions)
        hyper_factor = self._sparse_param.get("hyper_factor", 1.0)
        self._enumerate_function = HyperbolicAnisotropicEnumerateFunction(
            self._input_dim, hyper_factor
        )
        self._basis = self._get_basis()
        self._n_basis = self._get_basis_size()
        self._trunc_strategy = self._get_trunc_strategy()
        if self._strategy == self.QUAD_STRATEGY:
            (
                self._sample,
                self._weights,
                self._proj_strategy,
            ) = self._get_quadrature_points()
        else:
            self._sample = None
            self._weights = None

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        self._proj_strategy = self._get_proj_strategy(input_data, output_data)
        weights = self._get_weights(input_data)
        self.algo = self._build_pce(input_data, weights, output_data)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        return array(self.algo.getMetaModel()(input_data))

    @property
    def first_sobol_indices(self):  # type: (...) -> Dict[str,ndarray]
        """The first Sobol' indices."""
        sensitivity_analysis = FunctionalChaosSobolIndices(self.algo)
        LOGGER.info(str(sensitivity_analysis))
        first_order = {
            name: sensitivity_analysis.getSobolIndex(index)
            for index, name in enumerate(self.input_names)
        }
        return first_order

    @property
    def total_sobol_indices(self):  # type: (...) -> Dict[str,ndarray]
        """The total Sobol' indices."""
        sensitivity_analysis = FunctionalChaosSobolIndices(self.algo)
        LOGGER.info(str(sensitivity_analysis))
        total_order = {
            name: sensitivity_analysis.getSobolTotalIndex(index)
            for index, name in enumerate(self.input_names)
        }
        return total_order

    def _build_pce(
        self,
        input_data,  # type: ndarray
        weights,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> openturns.FunctionalChaosResult
        """Build the PCE with OpenTURNS.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).
            weights: The data weights.
            output_data: The output data with shape (n_samples, n_outputs).

        Returns:
            An OpenTURNS PCE.
        """
        pce_algo = FunctionalChaosAlgorithm(
            input_data, weights, output_data, self._dist, self._trunc_strategy
        )
        pce_algo.setProjectionStrategy(self._proj_strategy)
        pce_algo.run()
        return pce_algo.getResult()

    def _get_basis(self):  # type: (...) -> openturns.OrthogonalProductPolynomialFactory
        """Return the orthogonal product polynomial factory for PCE construction.

        Returns:
            An orthogonal product polynomial factory computed by OpenTURNS.
        """
        if self._stieltjes:
            # Tend to result in performance issue
            basis = OrthogonalProductPolynomialFactory(
                [
                    StandardDistributionPolynomialFactory(
                        AdaptiveStieltjesAlgorithm(marginal)
                    )
                    for marginal in self._ot_distributions
                ],
                self._enumerate_function,
            )
        else:
            basis = OrthogonalProductPolynomialFactory(
                [
                    StandardDistributionPolynomialFactory(margin)
                    for margin in self._ot_distributions
                ],
                self._enumerate_function,
            )
        return basis

    def _get_basis_size(self):  # type: (...) -> int
        """Return the basis size for PCE construction.

        Returns:
            The number of basis functions.
        """
        return self._enumerate_function.getStrataCumulatedCardinal(self._degree)

    def _get_quadrature_points(
        self,
    ):  # type: (...) -> Tuple[ndarray,ndarray,openturns.IntegrationStrategy]
        """Return the quadrature points for PCE construction.

        Returns:
            The quadrature points.
            The weights associated with the quadrature points.
            The projection strategy.
        """
        measure = self._basis.getMeasure()
        if self._n_quad is not None:
            degree_by_dim = int(self._n_quad ** (1.0 / self._input_dim))
        else:
            degree_by_dim = self._degree + 1
        degrees = [degree_by_dim] * self._input_dim

        proj_strategy = IntegrationStrategy(GaussProductExperiment(measure, degrees))
        sample, weights = proj_strategy.getExperiment().generateWithWeights()

        if not self._stieltjes:
            transformation = Function(
                MarginalTransformationEvaluation(
                    [measure.getMarginal(i) for i in range(self._input_dim)],
                    self._ot_distributions,
                    False,
                )
            )
            sample = transformation(sample)
        sample = array(sample)
        if not self.learning_set:
            inputs_names = self._prob_space.variables_names
            in_grp = self.learning_set.INPUT_GROUP
            self.learning_set.set_from_array(
                sample,
                self._prob_space.variables_names,
                self._prob_space.variables_sizes,
                {name: in_grp for name in inputs_names},
            )
        if self._discipline is not None:
            n_samples = len(self.learning_set)
            outputs = {name: [] for name in self._discipline.get_output_data_names()}
            for data in self.learning_set:
                output_data = self._discipline.execute(data)
                for name in self._discipline.get_output_data_names():
                    outputs[name] += list(output_data[name])
            for name in self._discipline.get_output_data_names():
                outputs[name] = array(outputs[name]).reshape((n_samples, -1))
            outputs_names = list(self._discipline.get_output_data_names())
            outputs = [outputs[name] for name in outputs_names]
            data = concatenate(outputs, axis=1)
            out_grp = self.learning_set.OUTPUT_GROUP
            sizes = {
                name: len(self._discipline.local_data[name]) for name in outputs_names
            }
            self.learning_set.add_group(
                out_grp, data, outputs_names, sizes, cache_as_input=False
            )
            self.output_names = outputs_names

        return sample, weights, proj_strategy

    def _get_trunc_strategy(
        self,
    ):  # type: (...) -> openturns.AdaptiveStrategyImplementation
        """Return the truncation strategy for PCE construction.

        Returns:
            The OpenTURNS truncation strategy.

        Raises:
            ValueError: If the truncation strategy is invalid.
        """
        if self._strategy == self.SPARSE_STRATEGY:
            sparse_param = self._sparse_param
            sparse_param = {} if sparse_param is None else sparse_param
            max_considered_terms = sparse_param.get("max_considered_terms", 120)
            most_significant = sparse_param.get("most_significant", 30)
            significance_factor = sparse_param.get("significance_factor", 1e-3)

            trunc_strategy = CleaningStrategy(
                OrthogonalBasis(self._basis),
                max_considered_terms,
                most_significant,
                significance_factor,
                True,
            )
        else:
            trunc_strategy = FixedStrategy(self._basis, self._n_basis)
        return trunc_strategy

    def _get_proj_strategy(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> openturns.LeastSquaresStrategy
        """Return the projection strategy for PCE construction.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).
            output_data: The output data with shape (n_samples, n_outputs).

        Returns:
            A projection strategy.
        """
        if self._strategy == self.QUAD_STRATEGY:
            proj_strategy = self._proj_strategy
        elif self._strategy == self.LS_STRATEGY:
            proj_strategy = LeastSquaresStrategy(input_data, output_data)
        else:
            app = LeastSquaresMetaModelSelectionFactory(LARS(), CorrectedLeaveOneOut())
            proj_strategy = LeastSquaresStrategy(input_data, output_data, app)
        return proj_strategy

    def _get_ls_weights(self):  # type: (...) -> openturns.Point
        """Return LS weights for PCE construction.

        Returns:
            The least-squares weights.
        """
        _, weights = self._proj_strategy.getExperiment().generateWithWeights()
        return weights

    def _get_quad_weights(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Return quadrature weights for PCE construction.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The weights.
        """
        sample = zeros_like(self._sample)
        common_len = len(input_data)
        sample[:common_len] = input_data
        sample[0] = self._sample[0]
        sample_arg = np_all(isin(sample, self._sample), axis=1)
        new_weights = array(self._weights)[sample_arg]
        return new_weights

    def _get_weights(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Return the weights for PCE construction.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).
        """
        if self._strategy == self.QUAD_STRATEGY:
            weights = self._get_quad_weights(input_data)
        else:
            weights = self._get_ls_weights()
        return weights

    def _predict_jacobian(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        input_shape, output_shape = self._get_raw_shapes()
        gradient = self.algo.getMetaModel().gradient
        jac = zeros((input_data.shape[0], int(output_shape), int(input_shape)))
        for index, data in enumerate(input_data):
            jac[index] = array(gradient(Point(data))).T
        return jac



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Introduction
************

Here is the Application Programming Interface (API) of |g|,
a set of high level functions for ease of use.

Make |g| ever more accessible
-----------------------------

The aim of this API is to provide high level functions that are sufficient
to use |g| in most cases, without requiring a deep knowledge of |g|.

Besides, these functions shall change much less often than the internal
classes, which is key for backward compatibility,
which means ensuring that your current scripts using |g| will be usable
with the future versions of |g|.

Connect |g| to your favorite tools
----------------------------------

The API also facilitates the interfacing of |g|
with a platform or other software.

To interface a simulation software with |g|,
please refer to: :ref:`software_connection`.

.. _extending-gemseo:

Extending |g|
-------------

|g| features can be extended with external python modules. All kinds of
additional features can be implemented: disciplines, algorithms, formulations,
post-processings, surrogate models, ... There are 2 ways to extend |g| with a
directory that contains the python modules:

- by adding the directory to the :envvar:`PYTHONPATH` if the directory name
  starts with :file:`gemseo_`,
- by setting the environment variable :envvar:`GEMSEO_PATH` to the directory
  path, multiple directories can be separated by :envvar:`:`.

Table of contents
*****************

Algorithms
----------

- :meth:`~gemseo.api.create_scenario`
- :meth:`~gemseo.api.execute_algo`
- :meth:`~gemseo.api.get_available_opt_algorithms`
- :meth:`~gemseo.api.get_available_doe_algorithms`
- :meth:`~gemseo.api.get_algorithm_options_schema`

Cache
-----

- :meth:`~gemseo.api.create_cache`
- :meth:`~gemseo.api.get_available_caches`

Configuration
-------------

- :meth:`~gemseo.api.configure_logger`

Coupling
--------

- :meth:`~gemseo.api.generate_n2_plot`
- :meth:`~gemseo.api.generate_coupling_graph`
- :meth:`~gemseo.api.get_all_inputs`
- :meth:`~gemseo.api.get_all_outputs`

Design space
------------

- :meth:`~gemseo.api.read_design_space`
- :meth:`~gemseo.api.export_design_space`
- :meth:`~gemseo.api.create_design_space`

Disciplines
-----------

- :meth:`~gemseo.api.create_discipline`
- :meth:`~gemseo.api.get_available_disciplines`
- :meth:`~gemseo.api.get_discipline_inputs_schema`
- :meth:`~gemseo.api.get_discipline_outputs_schema`
- :meth:`~gemseo.api.get_discipline_options_schema`
- :meth:`~gemseo.api.get_discipline_options_defaults`

Formulations
------------

- :meth:`~gemseo.api.create_scenario`
- :meth:`~gemseo.api.get_available_formulations`
- :meth:`~gemseo.api.get_formulation_options_schema`
- :meth:`~gemseo.api.get_formulation_sub_options_schema`
- :meth:`~gemseo.api.get_formulations_sub_options_defaults`
- :meth:`~gemseo.api.get_formulations_options_defaults`

MDA
---

- :meth:`~gemseo.api.create_mda`
- :meth:`~gemseo.api.get_available_mdas`
- :meth:`~gemseo.api.get_mda_options_schema`

Post-processing
---------------

- :meth:`~gemseo.api.execute_post`
- :meth:`~gemseo.api.get_available_post_processings`
- :meth:`~gemseo.api.get_post_processing_options_schema`

Scalable
--------

- :meth:`~gemseo.api.create_scalable`

Scenario
--------

- :meth:`~gemseo.api.create_scenario`
- :meth:`~gemseo.api.monitor_scenario`
- :meth:`~gemseo.api.get_available_scenario_types`
- :meth:`~gemseo.api.get_scenario_options_schema`
- :meth:`~gemseo.api.get_scenario_inputs_schema`
- :meth:`~gemseo.api.get_scenario_differenciation_modes`

Surrogates
----------

- :meth:`~gemseo.api.create_surrogate`
- :meth:`~gemseo.api.get_available_surrogates`
- :meth:`~gemseo.api.get_surrogate_options_schema`

API functions
*************
"""
from __future__ import division, unicode_literals

import logging
import re
from typing import List, Optional  # noqa F401

from six import string_types

from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.third_party.prettytable import PrettyTable
from gemseo.utils.logging_tools import MultiLineFileHandler, MultiLineStreamHandler

# Most modules are imported directly in the methods, which adds a very small
# overhead, but prevents users from importing them from the API.
# All factories are Singletons which means that the scan for
# plugins is done once only

LOGGER = logging.getLogger(__name__)

# pylint: disable=import-outside-toplevel


def generate_n2_plot(
    disciplines,
    file_path="n2.pdf",
    show_data_names=True,
    save=True,
    show=False,
    figsize=(15, 10),
    open_browser=False,
):
    """Generate a N2 plot for the disciplines list.

    :param disciplines: List of disciplines to analyze.
    :type disciplines: list(MDODiscipline)
    :param file_path: File path of the figure.
    :type file_path: str
    :param show_data_names: If true, the names of the
        coupling data is shown
        otherwise, circles are drawn, which size depend on the
        number of coupling names.
    :type show_data_names: bool
    :param save: If True, saved the figure to file_path.
    :type save: bool
    :param show: If True, shows the plot.
    :type show: bool
    :param figsize: Size of the figure.
    :type figsize: tuple(float)
    :param open_browser: If True, open the browser and display the N2.
    :type open_browser: bool

    Examples
    --------
    >>> from gemseo.api import create_discipline, generate_n2_plot
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> generate_n2_plot(disciplines)

    See also
    --------
    generate_coupling_graph
    get_all_inputs
    get_all_outputs
    """
    from gemseo.core.coupling_structure import MDOCouplingStructure

    coupling_structure = MDOCouplingStructure(disciplines)
    coupling_structure.plot_n2_chart(
        file_path, show_data_names, save, show, figsize, open_browser
    )


def generate_coupling_graph(disciplines, file_path="coupling_graph.pdf", full=True):
    """Generate a graph of the couplings for the disciplines list.

    :param disciplines: List of disciplines to analyze.
    :type disciplines: list(MDODiscipline)
    :param file_path: File path of the figure.
    :type file_path: str
    :param full: If True, generate the full coupling graph.
        Otherwise, generate the condensed one.
    :type full: bool

    Examples
    --------
    >>> from gemseo.api import create_discipline, generate_coupling_graph
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> generate_coupling_graph(disciplines)

    See also
    --------
    generate_n2_plot
    get_all_inputs
    get_all_outputs
    """
    from gemseo.core.coupling_structure import MDOCouplingStructure

    coupling_structure = MDOCouplingStructure(disciplines)
    if full:
        coupling_structure.graph.export_initial_graph(file_path)
    else:
        coupling_structure.graph.export_reduced_graph(file_path)


def get_available_formulations():
    """List the available formulations in the current configuration.

    :return: list of available MDO formulations.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_formulations
    >>> get_available_formulations()

    See also
    --------
    create_scenario
    get_formulation_options_schema
    get_formulation_sub_options_schema
    get_formulations_options_defaults
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    return MDOFormulationsFactory().formulations


def get_available_opt_algorithms():
    """List the available optimization algorithms names in the present configuration.

    :return: list of available optimization algorithms.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_opt_algorithms
    >>> get_available_opt_algorithms()

    See also
    --------
    create_scenario
    execute_algo
    get_available_doe_algorithms
    get_algorithm_options_schema
    """
    from gemseo.algos.opt.opt_factory import OptimizersFactory

    return OptimizersFactory().algorithms


def get_available_doe_algorithms():
    """List the available Design of Experiments algorithms names in the present
    configuration.

    :return: list of available DOE algorithms.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_doe_algorithms
    >>> get_available_doe_algorithms()

    See also
    --------
    create_scenario
    execute_algo
    get_available_opt_algorithms
    get_algorithm_options_schema
    """
    from gemseo.algos.doe.doe_factory import DOEFactory

    return DOEFactory().algorithms


def get_available_surrogates():
    """List the available surrogate model names in the present configuration.

    :return: list of available surrogate disciplines.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_surrogates
    >>> print(get_available_surrogates())
    ['RBFRegression', 'GaussianProcessRegression', 'LinearRegression', 'PCERegression']

    See also
    --------
    create_surrogate
    get_surrogate_options_schema
    """
    from gemseo.mlearning.api import get_regression_models

    return get_regression_models()


def get_available_disciplines():
    """List the available disciplines names in the present configuration.

    :return: list of available disciplines.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_disciplines
    >>> print(get_available_disciplines())
    ['RosenMF', 'SobieskiAerodynamics', 'ScalableKriging', 'DOEScenario',
    'MDOScenario', 'SobieskiMission', 'SobieskiBaseWrapper', 'Sellar1',
    'Sellar2', 'MDOChain', 'SobieskiStructure', 'AutoPyDiscipline',
    'Structure', 'SobieskiPropulsion', 'Scenario', 'AnalyticDiscipline',
    'MDOScenarioAdapter', 'ScalableDiscipline', 'SellarSystem', 'Aerodynamics',
    'Mission', 'PropaneComb1', 'PropaneComb2', 'PropaneComb3',
    'PropaneReaction', 'MDOParallelChain']

    See also
    --------
    create_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    return DisciplinesFactory().disciplines


def get_surrogate_options_schema(surrogate_name, output_json=False, pretty_print=False):
    """Lists the available options for a surrogate discipline.

    :param surrogate_name: Name of the surrogate discipline
    :type surrogate_name: str
    :param output_json: Apply json format for the schema
    :type output_json: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :returns: Option schema (string) of the surrogate discipline

    Examples
    --------
    >>> from gemseo.api import get_surrogate_options_schema
    >>> tmp = get_surrogate_options_schema('LinRegSurrogateDiscipline',
    >>>                                    pretty_print=True)

    See also
    --------
    create_surrogate
    get_available_surrogates
    """
    from gemseo.mlearning.api import get_regression_options

    return get_regression_options(surrogate_name, output_json, pretty_print)


def get_algorithm_options_schema(algorithm_name, output_json=False, pretty_print=False):
    """Get the options schema as a JSON Schema string or dictionary for a given
    algorithm.

    :param algorithm_name: Name of the algorithm
    :type algorithm_name: str
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: print the schema in a pretty table.
    :type pretty_print: bool
    :return: the option schema (string) of the algorithm

    Examples
    --------
    >>> from gemseo.api import get_algorithm_options_schema
    >>> schema = get_algorithm_options_schema('NLOPT_SLSQP', pretty_print=True)

    See also
    --------
    create_scenario
    execute_algo
    get_available_opt_algorithms
    get_available_doe_algorithms
    get_algorithm_options_schema
    """
    from gemseo.algos.doe.doe_factory import DOEFactory
    from gemseo.algos.opt.opt_factory import OptimizersFactory

    for factory in (DOEFactory(), OptimizersFactory()):
        if factory.is_available(algorithm_name):
            algo_lib = factory.create(algorithm_name)
            opts_gram = algo_lib.init_options_grammar(algorithm_name)
            schema = _get_schema(opts_gram, output_json, pretty_print)
            return schema
    raise ValueError("Algorithm named " + str(algorithm_name) + " is not available.")


def get_discipline_inputs_schema(discipline, output_json=False, pretty_print=False):
    """Get the inputs schema of a discipline.

    :param discipline: MDODiscipline instance.
    :type discipline: MDODiscipline
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :return: Option schema of the discipline inputs.
    :rtype: string

    Examples
    --------
    >>> from gemseo.api import create_discipline, get_discipline_inputs_schema
    >>> discipline = create_discipline('Sellar1')
    >>> schema = get_discipline_inputs_schema(discipline, pretty_print=True)

    See also
    --------
    create_discipline
    get_available_disciplines
    get_discipline_outputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    return _get_schema(discipline.input_grammar, output_json, pretty_print)


def get_discipline_outputs_schema(discipline, output_json=False, pretty_print=False):
    """Get the outputs schema of a discipline.

    :param discipline: MDODiscipline instance
    :type discipline: MDODiscipline
    :param output_json: If True, return a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :return: Schema of the discipline outputs
    :rtype: string

    Examples
    --------
    >>> from gemseo.api import get_discipline_outputs_schema, create_discipline
    >>> discipline = create_discipline('Sellar1')
    >>> get_discipline_outputs_schema(discipline, pretty_print=True)

    See also
    --------
    create_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    return _get_schema(discipline.output_grammar, output_json, pretty_print)


def get_available_post_processings():
    """List the available optimization post-processings.

    :return: List of available post-processings.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_post_processings
    >>> print(get_available_post_processings())
    ['ScatterPlotMatrix', 'VariableInfluence', 'ConstraintsHistory',
    'RadarChart', 'Robustness', 'Correlations', 'SOM', 'KMeans',
    'ParallelCoordinates', 'GradientSensitivity', 'OptHistoryView',
    'BasicHistory', 'ObjConstrHist', 'QuadApprox']

    See also
    --------
    execute_post
    get_post_processing_options_schema
    """
    from gemseo.post.post_factory import PostFactory

    return PostFactory().posts


def get_post_processing_options_schema(
    post_proc_name, output_json=False, pretty_print=False
):
    """List the options schema of a post-processing.

    :param post_proc_name: Post processing name.
    :type post_proc_name: str
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :type output_json: bool
    :return: Option schema of the post processing

    Examples
    --------
    >>> from gemseo.api import get_post_processing_options_schema
    >>> schema = get_post_processing_options_schema('OptHistoryView',
    >>>                                             pretty_print=True)

    See also
    --------
    execute_post
    get_available_post_processings
    """
    from gemseo.algos.opt_problem import OptimizationProblem
    from gemseo.post.post_factory import PostFactory

    post_proc = PostFactory().create(OptimizationProblem(None), post_proc_name)
    return _get_schema(post_proc.opt_grammar, output_json, pretty_print)


def get_formulation_options_schema(
    formulation_name, output_json=False, pretty_print=False
):
    """Get the options schema of a MDO formulation.

    :param formulation_name: Name of the MDO formulation.
    :type formulation_name: str
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :type output_json: bool
    :return: Option schema (string) of the MDO formulation.

    Examples
    --------
    >>> from gemseo.api import get_formulation_options_schema
    >>> schema = get_formulation_options_schema('MDF', pretty_print=True)

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_sub_options_schema
    get_formulations_options_defaults
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    grammar = factory.get_options_grammar(formulation_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_formulation_sub_options_schema(
    formulation_name, output_json=False, pretty_print=False, **formulation_options
):
    """Get the sub-options schema of a MDO formulation.

    :param formulation_name: Name of the MDO formulation
    :type formulation_name: str
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :param formulation_options: Options to be passed to the formulation;
        this is required to instantiate it.
    :return: Sub-option schema of the MDO formulation

    Examples
    --------
    >>> from gemseo.api import get_formulation_sub_options_schema
    >>> schema = get_formulation_sub_options_schema('MDF',
    >>>                                             main_mda_class='MDAJacobi',
    >>>                                             pretty_print=True)

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_options_schema
    get_formulations_options_defaults
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    grammar = factory.get_sub_options_grammar(formulation_name, **formulation_options)
    return _get_schema(grammar, output_json, pretty_print)


def get_formulations_sub_options_defaults(formulation_name, **formulation_options):
    """Get the default values of the sub options of a formulation.

    :param formulation_name: Name of the discipline.
    :type formulation_name: str
    :param formulation_options: Options to be passed to the formulation;
        this is required to instantiate it.
    :return: Default values of the sub-options of the MDO formulation.

    Examples
    --------
    >>> from gemseo.api import get_formulations_sub_options_defaults
    >>> get_formulations_sub_options_defaults('MDF',
    >>>                                       main_mda_class='MDAJacobi')

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_options_schema
    get_formulation_sub_options_schema
    get_formulations_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    return factory.get_default_sub_options_values(
        formulation_name, **formulation_options
    )


def get_formulations_options_defaults(formulation_name):
    """Get the default values of the options of a formulation.

    :param formulation_name: Name of the discipline.
    :type formulation_name: str
    :return: Default values of the options of the MDO formulation.

    Examples
    --------
    >>> from gemseo.api import get_formulations_options_defaults
    >>> get_formulations_options_defaults('MDF')
    {'main_mda_class': 'MDAChain',
     'maximize_objective': False,
     'sub_mda_class': 'MDAJacobi'}

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_options_schema
    get_formulation_sub_options_schema
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    return factory.get_default_options_values(formulation_name)


def get_discipline_options_schema(
    discipline_name, output_json=False, pretty_print=False
):
    """Get the options schema of a discipline.

    :param discipline_name: Name of the discipline.
    :type discipline_name: str
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: print the schema in a pretty table.
    :type pretty_print: bool
    :return: Options schema of the discipline

    Examples
    --------
    >>> from gemseo.api import get_discipline_options_schema
    >>> schema = get_discipline_options_schema('Sellar1', pretty_print=True)

    See also
    --------
    create_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_defaults
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    disc_fact = DisciplinesFactory()
    grammar = disc_fact.get_options_grammar(discipline_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_scenario_options_schema(scenario_type, output_json=False, pretty_print=False):
    """Get the options schema of a scenario.

    :param scenario_type: Type of scenario (DOE, MDO...)
    :type scenario_type: str
    :param output_json: True, returns a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :return: Options schema of the scenario

    Examples
    --------
    >>> from gemseo.api import get_scenario_options_schema
    >>> get_scenario_options_schema('MDO')

    See also
    --------
    create_scenario
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differenciation_modes
    """
    if scenario_type not in get_available_scenario_types():
        raise ValueError("Unknown scenario type " + str(scenario_type))
    scenario_class = {"MDO": "MDOScenario", "DOE": "DOEScenario"}[scenario_type]
    return get_discipline_options_schema(scenario_class, output_json, pretty_print)


def get_scenario_inputs_schema(scenario, output_json=False, pretty_print=False):
    """Get the schema of the inputs of a scenario.

    :param scenario: Scenario instance
    :type scenario: Scenario
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :return: Schema of the scenario inputs

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_scenario,
    get_scenario_inputs_schema
    >>> from gemseo.problems.sellar.sellar_design_space import SellarDesignSpace
    >>> design_space = SellarDesignSpace()
    >>> disciplines = create_discipline(['Sellar1','Sellar2','SellarSystem'])
    >>> scenario = create_scenario(disciplines, 'MDF', 'obj', design_space,
    'my_scenario', 'MDO')
    >>> get_scenario_inputs_schema(scenario)

    See also
    --------
    create_scenario
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_differenciation_modes
    """
    return get_discipline_inputs_schema(scenario, output_json, pretty_print)


def get_discipline_options_defaults(discipline_name):
    """Get the default values of the options of a discipline.

    :param discipline_name: Name of the discipline.
    :type discipline_name: str
    :return: Default option values of a discipline.

    Examples
    --------
    >>> from gemseo.api import get_discipline_options_defaults
    >>> get_discipline_options_defaults('Sellar1')

    See also
    --------
    create_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_schema
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    factory = DisciplinesFactory().factory
    return factory.get_default_options_values(discipline_name)


def get_scenario_differenciation_modes():
    """List the available differenciation modes of a scenario.

    :returns: List of differenciation modes.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_scenario_differenciation_modes
    >>> get_scenario_differenciation_modes()

    See also
    --------
    create_scenario
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    """
    from gemseo.algos.opt_problem import OptimizationProblem

    return OptimizationProblem.DIFFERENTIATION_METHODS


def get_available_scenario_types():
    """List the available scenario types.

    :return: list of available scenario types.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_scenario_types
    >>> get_available_scenario_types()

    See also
    --------
    create_scenario
    monitor_scenario
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differenciation_modes
    """
    return ["MDO", "DOE"]


def _get_schema(json_grammar, output_json=False, pretty_print=False):
    """Get the schema of a JSON grammar.

    :param json_grammar: a JSONGrammar instance
    :type json_grammar: JSONGrammar
    :param output_json: if True, returns a JSON string,
    :type output_json: bool
    :param pretty_print: print the schema in a pretty table.
    :type pretty_print: bool
    :return: a schema
    :rtype: dict or str
    """
    if json_grammar is None:
        return None
    schema = json_grammar.schema
    dict_schema = schema.to_dict()
    if pretty_print:
        if "name" in dict_schema:
            title = dict_schema["name"].replace("_", " ")
        else:
            title = None
        table = PrettyTable(title=title, max_table_width=150)
        names = []
        descriptions = []
        types = []
        for name, value in dict_schema["properties"].items():
            names.append(name)
            descriptions.append(value.get("description"))
            description = descriptions[-1]
            tmp = []
            if descriptions[-1] is not None:
                descriptions[-1] = descriptions[-1].split(":type")[0]
                descriptions[-1] = descriptions[-1].capitalize()
                descriptions[-1] = descriptions[-1].replace("\n", " ")
                tmp = re.split(r":type ([*\w]+): (.*?)", description)
            if len(tmp) == 4:
                types.append(tmp[3].strip())
            else:
                types.append(value.get("type"))
        table.add_column("Name", names)
        table.add_column("Description", descriptions)
        table.add_column("Type", types)
        table.sortby = "Name"
        table.min_width = 25
    if output_json:
        return schema.to_json()
    return dict_schema


def get_available_mdas():
    """List the available MDAs.

    :return: list of available MDAs.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_mdas
    >>> get_available_mdas()

    See also
    --------
    create_mda
    get_mda_options_schema
    """
    from gemseo.mda.mda_factory import MDAFactory

    return MDAFactory().mdas


def get_mda_options_schema(mda_name, output_json=False, pretty_print=False):
    """Get the options schema of a MDA.

    :param mda_name: Name of the MDA.
    :type mda_name: str
    :param output_json: If True, returns a JSON string,
        return a dict otherwise.
    :type output_json: bool
    :param pretty_print: Print the schema in a pretty table.
    :type pretty_print: bool
    :return: MDA options schema
    :rtype: dict or str

    Examples
    --------
    >>> from gemseo.api import get_mda_options_schema
    >>> get_mda_options_schema('MDAJacobi')

    See also
    --------
    create_mda
    get_available_mdas
    """
    from gemseo.mda.mda_factory import MDAFactory

    factory = MDAFactory().factory
    grammar = factory.get_options_grammar(mda_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_all_inputs(disciplines, recursive=False):
    """List all the inputs of the disciplines.

    Merge the input data from the disciplines grammars.

    :param disciplines: List of disciplines to search
        for inputs.
    :type disciplines: list(MDODiscipline)
    :param recursive: If True, searches for the inputs of the
        sub disciplines (when some disciplines are scenarios).
    :type recursive: bool
    :returns: List of input data.

    Examples
    --------
    >>> from gemseo.api import create_discipline, get_all_inputs
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2'])
    >>> get_all_inputs(disciplines)
    ['y_0', 'x_shared', 'y_1', 'x_local']

    See also
    --------
    generate_n2_plot
    generate_coupling_graph
    get_all_outputs
    """
    from gemseo.utils.data_conversion import DataConversion

    return DataConversion.get_all_inputs(disciplines, recursive)


def get_all_outputs(disciplines, recursive=False):
    """List all the outputs of the disciplines.

    Merge the output data from the disicplines grammars.

    :param disciplines: List of disciplines to search
        for outputs.
    :type disciplines: list(MDODiscipline)
    :param recursive: If True, searches for the outputs of the
        sub disciplines (when some disciplines are scenarios).
    :type recursive: bool
    :returns: List of input data

    Examples
    --------
    >>> from gemseo.api import create_discipline, get_all_outputs
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2'])
    >>> get_all_outputs(disciplines)
    ['y_1', 'y_0']

    See also
    --------
    generate_n2_plot
    generate_coupling_graph
    get_all_inputs
    """
    from gemseo.utils.data_conversion import DataConversion

    return DataConversion.get_all_outputs(disciplines, recursive)


def create_scenario(
    disciplines,
    formulation,
    objective_name,
    design_space,
    name=None,
    scenario_type="MDO",
    maximize_objective=False,
    **options
):
    """Create a scenario.

    :param disciplines: Disciplines of the scenario.
    :type disciplines: list(MDODiscipline)
    :param formulation: Formulation name.
    :type formulation: str
    :param objective_name: Objective function name.
    :type objective_name: str
    :param design_space: Design space object or a file that contains
        the design space.
    :type design_space: DesignSpace or str
    :param name: scenario name
    :type name: str
    :param scenario_type: Type of scenario, e.g. "DOE" or "MDO".
    :type scenario_type: str
    :param maximize_objective: Maximize function objective.
    :type maximize_objective: bool
    :param options: Options passed to the MDO formulation.
    :returns: Scenario.
    :rtype: Scenario

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_scenario
    >>> from gemseo.problems.sellar.sellar_design_space import SellarDesignSpace
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> design_space = SellarDesignSpace()
    >>> scenario = create_scenario(disciplines, 'MDF', 'obj', design_space,
    >>>                            'SellarMDFScenario', 'MDO')

    See also
    --------
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differenciation_modes
    """
    from gemseo.core.doe_scenario import DOEScenario
    from gemseo.core.mdo_scenario import MDOScenario

    if not isinstance(disciplines, list):
        disciplines = [disciplines]

    if isinstance(design_space, string_types):
        design_space = read_design_space(design_space)

    if scenario_type == "MDO":
        return MDOScenario(
            disciplines,
            formulation,
            objective_name,
            design_space,
            name,
            maximize_objective=maximize_objective,
            **options
        )

    if scenario_type == "DOE":
        return DOEScenario(
            disciplines,
            formulation,
            objective_name,
            design_space,
            name,
            maximize_objective=maximize_objective,
            **options
        )

    raise ValueError(
        "Unknown scenario type :"
        + str(scenario_type)
        + ", use one of : 'MDO' or 'DOE'."
    )


def configure_logger(
    logger_name=None,
    level=logging.INFO,
    date_format="%H:%M:%S",
    message_format="%(levelname)8s - %(asctime)s: %(message)s",
    filename=None,
    filemode="a",
):
    """Configure |g| logging.

    :param logger_name: Name of the logger to configure, default: None, i.e.
        the root logger.
    :type logger_name: str
    :param level: Logging level among 'DEBUG', 'INFO', 'WARNING' or
        'CRITICAL's, default: 'INFO'.
    :type level: str
    :param date_format: Logging date format, default: %H:%M:%S.
    :type date_format: str
    :param message_format: logging message format, default: %(levelname)8s -
        %(asctime)s: %(message)s.
    :type message_format: str
    :param filename: File path if outputs must be written in a file, default: None.
    :type filename: str
    :param filemode: Logging output file mode, 'w' (overwrite) or 'a' (append),
        default: 'a'.
    :type filemode: str

    Examples
    --------
    >>> import logging
    >>> configure_logger(logging.WARNING)
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    formatter = logging.Formatter(fmt=message_format, datefmt=date_format)
    # remove all existing handlers
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)
    stream_handler = MultiLineStreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    if filename is not None:
        file_handler = MultiLineFileHandler(
            filename, mode=filemode, delay=True, encoding="utf-8"
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return logger


def create_discipline(discipline_name, **options):
    """Create disciplines that are known to |g|.

    |g| knows the disciplines located in the following directories:

    - the directories listed in the environment variable *GEMSEO_PATH*,
    - the directories located in the *problems* package,

    :param discipline_name: Name of the discipline, or list of names.
    :type discipline_name: MDODiscipline
    :param options: Additional options to be passed to the discipline
        constructor.
    :returns: Disciplines.
    :rtype: list(MDODiscipline)

    Examples
    --------
    >>> from gemseo.api import create_discipline
    >>> discipline = create_discipline('Sellar1')
    >>> discipline.execute()
    {'x_local': array([0.+0.j]),
     'x_shared': array([1.+0.j, 0.+0.j]),
     'y_0': array([0.89442719+0.j]),
     'y_1': array([1.+0.j])}

    See also
    --------
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    factory = DisciplinesFactory()
    if isinstance(discipline_name, list):
        return [factory.create(d_name, **options) for d_name in discipline_name]
    return factory.create(discipline_name, **options)


def create_scalable(name, data, sizes=None, **parameters):
    """Create a scalable discipline.

    :param str name: scalable model class name.
    :param Dataset data: learning dataset.
    :param dict sizes: sizes of input and output variables.
    :param parameters: model parameters
    """
    from gemseo.problems.scalable.data_driven.discipline import ScalableDiscipline

    return ScalableDiscipline(name, data, sizes, **parameters)


def create_surrogate(
    surrogate,
    data=None,
    transformer=MLRegressionAlgo.DEFAULT_TRANSFORMER,
    disc_name=None,
    default_inputs=None,
    input_names=None,
    output_names=None,
    **parameters
):
    """Create a surrogate discipline.

    :param surrogate: name of the surrogate model algorithm.
    :type surrogate: str or MLRegressionAlgo
    :param Dataset data: dataset to train the surrogate. If None,
        assumes that the surrogate is trained. Default: None.
    :param dict(Transformer) transformer: transformation strategy for data groups.
        If None, do not scale data.
        Default: MLRegressionAlgo.DEFAULT_SCALER,
        which is a min/max scaler applied to the inputs
        and a min/max scaler applied to the outputs.
    :param str disc_name: Surrogate discipline name.
    :param dict default_inputs: default inputs. If None, use the center of the
        learning input space. Default: None.
    :param list(str) input_names: list of input names. If None, use all inputs.
        Default: None.
    :param list(str) output_names: list of output names. If None, use all
        outputs. Default: None.
    :param parameters: Additional parameters to be passed to the surrogate
        for its construction.
    :returns: Surrogate discipline instance.
    :rtype: SurrogateDiscipline

    See also
    --------
    get_available_surrogates
    get_surrogate_options_schema
    """
    from gemseo.core.surrogate_disc import SurrogateDiscipline

    return SurrogateDiscipline(
        surrogate,
        data,
        transformer,
        disc_name,
        default_inputs,
        input_names,
        output_names,
        **parameters
    )


def create_mda(mda_name, disciplines, **options):
    """Create an MDA.

    :param mda_name: Name of the MDA (its classname).
    :type mda_name: str
    :param disciplines: List of the disciplines.
    :type disciplines: list(MDODiscipline)
    :param options: Additional options specific to the MDA.
    :returns: MDA instance.
    :rtype: MDA

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_mda
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2'])
    >>> mda = create_mda('MDAGaussSeidel', disciplines)
    >>> mda.execute()
    {'x_local': array([0.+0.j]),
     'x_shared': array([1.+0.j, 0.+0.j]),
     'y_0': array([0.79999995+0.j]),
     'y_1': array([1.79999995+0.j])}

    See also
    --------
    create_mda
    get_available_mdas
    get_mda_options_schema
    """
    from gemseo.mda.mda_factory import MDAFactory

    factory = MDAFactory()
    return factory.create(mda_name=mda_name, disciplines=disciplines, **options)


def execute_post(to_post_proc, post_name, **options):
    """Execute a post-processing method.

    :param to_post_proc: MDO or DOE scenario, or an optimization problem,
        or a path to a HDF file containing a saved OptimizationProblem.
    :type to_post_proc: MDOScenario, DOEScenario, OptimizationProblem,
        or str
    :param post_name: Post processing name.
    :type post_name: str
    :param options: Post-processing options.

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_scenario, execute_post
    >>> from gemseo.problems.sellar.sellar_design_space import SellarDesignSpace
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> design_space = SellarDesignSpace()
    >>> scenario = create_scenario(disciplines, 'MDF', 'obj', design_space,
    >>>                            'SellarMDFScenario', 'MDO')
    >>> scenario.execute({'algo': 'NLOPT_SLSQP', 'max_iter': 100})
    >>> execute_post(scenario, 'OptHistoryView', show=False, save=True)

    See also
    --------
    get_available_post_processings
    get_post_processing_options_schema
    """
    from gemseo.algos.opt_problem import OptimizationProblem
    from gemseo.post.post_factory import PostFactory

    if hasattr(to_post_proc, "is_scenario") and to_post_proc.is_scenario():
        opt_problem = to_post_proc.formulation.opt_problem
    elif isinstance(to_post_proc, OptimizationProblem):
        opt_problem = to_post_proc
    elif isinstance(to_post_proc, string_types):
        opt_problem = OptimizationProblem.import_hdf(to_post_proc)
    else:
        raise TypeError("Cannot post process type : " + str(type(to_post_proc)))
    return PostFactory().execute(opt_problem, post_name, **options)


def execute_algo(opt_problem, algo_name, algo_type="opt", **options):
    """Solve an optimization problem using either a DOE or an Optimization algorithm.

    :param opt_problem: the OptimizationProblem to be solved.
    :type opt_problem: OptimizationProblem
    :param algo_name: Name of the algorithm to be used to solve
        the problem.
    :type algo_name: str
    :param algo_type: "opt" or "doe" to use an optimization or a
        Design of Experiments algorithm
    :type algo_type: str
    :param options: algorithm options

    Examples
    --------
    >>> from gemseo.api import execute_algo
    >>> from gemseo.problems.analytical.rosenbrock import Rosenbrock
    >>> opt_problem = Rosenbrock()
    >>> opt_result = execute_algo(opt_problem, 'SLSQP')
    >>> opt_result
    Optimization result:
    |_ Design variables: [0.99999787 0.99999581]
    |_ Objective function: 5.054173713127532e-12
    |_ Feasible solution: True

    See also
    --------
    get_available_opt_algorithms
    get_available_doe_algorithms
    get_algorithm_options_schema
    """
    if algo_type == "opt":
        from gemseo.algos.opt.opt_factory import OptimizersFactory

        factory = OptimizersFactory()

    elif algo_type == "doe":
        from gemseo.algos.doe.doe_factory import DOEFactory

        factory = DOEFactory()
    else:
        raise ValueError(
            "Unknown algo type: " + str(algo_type) + ", please use 'doe' or 'opt' !"
        )

    return factory.execute(opt_problem, algo_name, **options)


def monitor_scenario(scenario, observer):
    """Adds an observer to a scenario The observer must have an "update(atom)" method
    that handles the execution status change of atom update(atom) is called everytime an
    atom execution changes.

    :param scenario: Scenario to monitor
    :type scenario: Scenario
    :param observer: Observer that handles an update of status

    See also
    --------
    create_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differenciation_modes
    """
    from gemseo.core.monitoring import Monitoring

    # Monitoring object is a singleton
    monitor = Monitoring(scenario)
    monitor.add_observer(observer)


def print_configuration():
    """Print the configuration with successfully loaded modules and failed imports with
    the reason.

    Examples
    --------
    >>> from gemseo.api import print_configuration
    >>> print_configuration()
    """
    from gemseo.algos.doe.doe_factory import DOEFactory
    from gemseo.algos.opt.opt_factory import OptimizersFactory
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory
    from gemseo.mda.mda_factory import MDAFactory
    from gemseo.mlearning.regression.factory import RegressionModelFactory
    from gemseo.post.post_factory import PostFactory
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    for factory in (
        DisciplinesFactory,
        OptimizersFactory,
        DOEFactory,
        RegressionModelFactory,
        MDOFormulationsFactory,
        MDAFactory,
        PostFactory,
    ):
        LOGGER.info("%s", factory().factory)


def read_design_space(file_path, header=None):
    """Read a file containing a design space and return a DesignSpace object.

    :param file_path: Path to the text file
        shall contain a CSV with a row for each variable
        and at least the bounds of the variable.
    :type file_path: str
    :param header: Fields list, or by default, read in the file
    :type header: list(str)
    :returns:  Design space
    :rtype: DesignSpace

    Examples
    --------
    >>> from gemseo.api import (create_design_space, export_design_space,
    >>>     read_design_space)
    >>> source_design_space = create_design_space()
    >>> source_design_space.add_variable('x', l_b=-1, value=0., u_b=1.)
    >>> export_design_space(source_design_space, 'file.txt')
    >>> read_design_space = read_design_space('file.txt')
    >>> print(read_design_space)
    Design Space:
    +------+-------------+-------+-------------+-------+
    | name | lower_bound | value | upper_bound | type  |
    +------+-------------+-------+-------------+-------+
    | x    |      -1     |   0   |      1      | float |
    +------+-------------+-------+-------------+-------+

    See also
    --------
    export_design_space
    create_design_space
    """
    from gemseo.algos.design_space import DesignSpace

    return DesignSpace.read_from_txt(file_path, header)


def export_design_space(
    design_space,
    output_file,
    export_hdf=False,
    fields=None,
    header_char="",
    **table_options
):
    """Export a design space to a text or HDF file.

    :param design_space: Design space.
    :type design_space: DesignSpace
    :param export_hdf: Export to a HDF file (True, default)
        or a txt file (False).
    :type export_hdf: bool
    :param output_file: Output file path.
    :type output_file: str
    :param fields: List of fields to export, by default all.
    :type fields: list(str)
    :param header_char: Header to use when exporting to a text file.
        Default: "".
    :type header_char: str

    Examples
    --------
    >>> from gemseo.api import create_design_space, export_design_space
    >>> design_space = create_design_space()
    >>> design_space.add_variable('x', l_b=-1, u_b=1, value=0.)
    >>> export_design_space(design_space, 'file.txt')

    See also
    --------
    read_design_space
    create_design_space
    """
    if export_hdf:
        design_space.export_hdf(output_file)
    else:
        design_space.export_to_txt(output_file, fields, header_char, **table_options)


def create_design_space():
    """Create an empty instance of a DesignSpace.

    :returns: Empty design space
    :rtype: DesignSpace

    Examples
    --------
    >>> from gemseo.api import create_design_space
    >>> design_space = create_design_space()
    >>> design_space.add_variable('x', l_b=-1, u_b=1, value=0.)
    >>> print(design_space)
    Design Space:
    +------+-------------+-------+-------------+-------+
    | name | lower_bound | value | upper_bound | type  |
    +------+-------------+-------+-------------+-------+
    | x    |      -1     |   0   |      1      | float |
    +------+-------------+-------+-------------+-------+

    See also
    --------
    read_design_space
    export_design_space
    create_design_space
    gemseo.algos.design_space.DesignSpace
    """
    from gemseo.algos.design_space import DesignSpace

    return DesignSpace()


def create_parameter_space():
    """Create an empty instance of a ParameterSpace.

    :returns: Empty parameter space
    :rtype: ParameterSpace
    """
    from gemseo.algos.parameter_space import ParameterSpace

    return ParameterSpace()


def get_available_caches():
    """Return available caches.

    :return: list of available caches.
    :rtype: list(str)

    Examples
    --------
    >>> from gemseo.api import get_available_caches
    >>> get_available_caches()
    ['AbstractFullCache', 'HDF5Cache', 'MemoryFullCache', 'SimpleCache']

    See also
    --------
    get_available_caches
    gemseo.core.discipline.MDODiscipline.set_cache_policy
    """
    from gemseo.caches.cache_factory import CacheFactory

    return CacheFactory().caches


def create_cache(cache_type, name=None, **options):
    """Return a cache.

    :param str cache_type: type of cache.
    :param str name: name of the cache. If None, use cache_type.
    :param options: options specific to cache_type

    Examples
    --------
    >>> from gemseo.api import create_cache
    >>> cache = create_cache('MemoryFullCache')
    >>> print(cache)
    +--------------------------------+
    |        MemoryFullCache         |
    +--------------+-----------------+
    |   Property   |      Value      |
    +--------------+-----------------+
    |     Type     | MemoryFullCache |
    |  Tolerance   |       0.0       |
    | Input names  |       None      |
    | Output names |       None      |
    |    Length    |        0        |
    +--------------+-----------------+

    See also
    --------
    get_available_caches
    gemseo.core.discipline.MDODiscipline.set_cache_policy
    """
    from gemseo.caches.cache_factory import CacheFactory

    return CacheFactory().create(cache_type, name=name, **options)


def create_dataset(
    name,
    data,
    variables=None,
    sizes=None,
    groups=None,
    by_group=True,
    delimiter=",",
    header=True,
    default_name=None,
):
    """Create a dataset from a numpy array or a file.

    :param str name: dataset name.
    :param data: array dataset or file path.
    :type data: array or str
    :param list(str) variables: list of variables names.
    :param dict(int) sizes: list of variables sizes.
    :param dict(str) groups: list of variables groups.
    :param bool by_group: if True, store the data by group. Otherwise,
        store them by variables. Default: True.
    :param str delimiter: field delimiter. Default: ','.
    :param bool header: if True and data is a string, read the variables names
        on the first line of the file. Default: True.
    :param str default_name: default variable name.

    See also
    --------
    load_dataset
    """
    from gemseo.core.dataset import Dataset

    dataset = Dataset(name, by_group)
    if isinstance(data, str):
        dataset.set_from_file(data, variables, sizes, groups, delimiter, header)
    else:
        dataset.set_from_array(data, variables, sizes, groups, default_name)
    return dataset


def load_dataset(dataset, **options):
    """Create a dataset from an existing subclass of Dataset. Typically, benchmark
    datasets can be found in gemseo.problems.dataset.

    :param str dataset: dataset name (its classname).

    See also
    --------
    create_dataset
    """
    from gemseo.problems.dataset.factory import DatasetFactory

    return DatasetFactory().create(dataset, **options)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Benoit Pauwels
"""Build matrices from linear constraints for solvers."""
from numpy import hstack, isfinite, vstack, zeros

from gemseo.core.function import MDOLinearFunction


def build_constraints_matrices(constraints, constraint_type):
    """Build the constraints matrices associated with passed linear constraints.

    :param constraints: list of linear constraints
    :type constraints: list(MDOLinearFunction)
    :param constraint_type: type of constraint to consider
    :type constraint_type: str
    :returns: left-hand side matrix, right-hand side vector
    :rtype: ndarray or None, ndarray or None
    """
    # Check the constraint type
    valid_types = [MDOLinearFunction.TYPE_INEQ, MDOLinearFunction.TYPE_EQ]
    if constraint_type not in valid_types:
        raise ValueError(
            "{} is not among valid constraint types {}".format(
                constraint_type, " ".join(valid_types)
            )
        )

    # Filter the constraints to consider
    cstr_list = [cstr for cstr in constraints if cstr.f_type == constraint_type]
    if not cstr_list:
        return None, None

    # Check that the constraint are linear
    for cstr in cstr_list:
        if not isinstance(cstr, MDOLinearFunction):
            raise TypeError(
                'Constraint "{}" is not an MDOLinearFunction'.format(cstr.name)
            )

    # Build the constraints matrices
    lhs_matrix = vstack([cstr.coefficients for cstr in cstr_list])
    rhs_vector = hstack([-cstr.value_at_zero for cstr in cstr_list])

    return lhs_matrix, rhs_vector


def build_bounds_matrices(bounds, upper):
    """Return the constraint matrices corresponding to bounds.

    :param bounds: value of the bounds
    :type bounds: ndarray
    :param upper: if True the bounds are considered upper bounds
    :type upper: bool
    :return: left-hand side matrix, right-hand side vector
    :rtype: ndarray, ndarray
    """
    is_finite = isfinite(bounds)
    n_finite = is_finite.sum()
    if n_finite == 0:
        return None, None
    lhs_mat = zeros((n_finite, bounds.size))
    lhs_mat[(range(n_finite), is_finite)] = 1.0 if upper else -1.0
    rhs_vec = bounds[is_finite] if upper else -bounds[is_finite]
    return lhs_mat, rhs_vec



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Capabilities to create and manipulate probability distributions.

This package contains:

- an abstract class :class:`.Distribution`
  to define the concept of probability distribution,
- an abstract class :class:`.ComposedDistribution`
  to define the concept of joint probability distribution
  by composing several instances of :class:`.Distribution`,
- a factory :class:`.DistributionFactory` to create instances of :class:`.Distribution`,
- concrete classes implementing these abstracts concepts, by interfacing:

  - the OpenTURNS library:
    :class:`.OTDistribution` and :class:`.OTComposedDistribution`,
  - the Scipy library:
    :class:`.SPDistribution` and :class:`.SPComposedDistribution`.

Lastly, the class :class:`.OTDistributionFitter` offers the possibility
to fit an :class:`.OTDistribution` from data based on OpenTURNS.
"""



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Some chains of SSBJ disciplines
*******************************
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.chain import MDOChain
from gemseo.mda.gauss_seidel import MDAGaussSeidel
from gemseo.mda.jacobi import MDAJacobi
from gemseo.problems.sobieski.wrappers import (
    SobieskiAerodynamics,
    SobieskiMission,
    SobieskiPropulsion,
    SobieskiStructure,
)

LOGGER = logging.getLogger(__name__)


class SobieskiChain(MDOChain):
    """Chains Sobieski disciplines : weight, aero, power and range"""

    def __init__(self, dtype="float64"):
        """Constructor.

        :param dtype: data array type, either "float64" or "complex128".
        :type dtype: str
        """
        disciplines = [
            SobieskiStructure(dtype),
            SobieskiAerodynamics(dtype),
            SobieskiPropulsion(dtype),
            SobieskiMission(dtype),
        ]
        super(SobieskiChain, self).__init__(disciplines, "SobieskiChain")


class SobieskiMDAGaussSeidel(MDAGaussSeidel):
    """Chains Sobieski disciplines to perform and MDA by Gauss Seidel algorithm Loops
    over Sobieski wrappers."""

    def __init__(self, dtype="float64", **mda_options):
        """Constructor of a MDA using Gauss-Seidel.

        :param dtype: data array type, either "float64" or "complex128".
        :type dtype: str
        :param mda_options: MDA options
        """
        disciplines = [
            SobieskiStructure(dtype),
            SobieskiAerodynamics(dtype),
            SobieskiPropulsion(dtype),
            SobieskiMission(dtype),
        ]
        super(SobieskiMDAGaussSeidel, self).__init__(disciplines, **mda_options)


class SobieskiMDAJacobi(MDAJacobi):
    """Chains Sobieski disciplines to perform and MDA by Jacobi algorithm Loops over
    Sobieski wrappers."""

    def __init__(self, n_processes=1, dtype="float64", **mda_options):
        """Constructor of a MDA using Jacobi.

        :param n_processes: maximum number of processors on which to run
        :type n_processes: integer
        :param dtype: data array type, either "float64" or "complex128".
        :type dtype: str
        :param mda_options: MDA options
        """
        disciplines = [
            SobieskiStructure(dtype),
            SobieskiAerodynamics(dtype),
            SobieskiPropulsion(dtype),
            SobieskiMission(dtype),
        ]
        super(SobieskiMDAJacobi, self).__init__(
            disciplines, n_processes=n_processes, **mda_options
        )
